# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import numpy as np
import scipy.sparse as sp
from scipy.linalg import hilbert
from pyomo.contrib.pynumero.linalg.mumps_interface import (
    MumpsCentralizedAssembledLinearSolver,
)


def main():
    # create the matrix and the right hand sides
    N = 1000
    A = sp.coo_matrix(
        hilbert(N) + np.identity(N)
    )  # a well-condition, symmetric, positive-definite matrix with off-diagonal entries
    true_x1 = np.arange(N)
    true_x2 = np.array(list(reversed(np.arange(N))))
    b1 = A * true_x1
    b2 = A * true_x2

    # solve
    solver = MumpsCentralizedAssembledLinearSolver()
    x1, res = solver.solve(A, b1)
    x2, res = solver.solve(A, b2)
    assert np.allclose(x1, true_x1)
    assert np.allclose(x2, true_x2)

    # only perform factorization once
    solver = MumpsCentralizedAssembledLinearSolver()
    solver.do_symbolic_factorization(A)
    solver.do_numeric_factorization(A)
    x1, res = solver.do_back_solve(b1)
    x2, res = solver.do_back_solve(b2)
    assert np.allclose(x1, true_x1)
    assert np.allclose(x2, true_x2)

    # Tell Mumps the matrix is symmetric
    # Note that the answer will be incorrect if both the lower
    # and upper portions of the matrix are given.
    solver = MumpsCentralizedAssembledLinearSolver(sym=2)
    A_lower_triangular = sp.tril(A)
    x1, res = solver.solve(A_lower_triangular, b1)
    assert np.allclose(x1, true_x1)

    # Tell Mumps the matrix is symmetric and positive-definite
    solver = MumpsCentralizedAssembledLinearSolver(sym=1)
    A_lower_triangular = sp.tril(A)
    x1, res = solver.solve(A_lower_triangular, b1)
    assert np.allclose(x1, true_x1)

    # Set options
    solver = MumpsCentralizedAssembledLinearSolver(
        icntl_options={11: 2}
    )  # compute error stats
    solver.set_cntl(2, 1e-4)  # set the stopping criteria for iterative refinement
    solver.set_icntl(
        10, 5
    )  # set the maximum number of iterations for iterative refinement to 5
    x1, res = solver.solve(A, b1)
    assert np.allclose(x1, true_x1)

    # Get information after the solve
    print(
        'Number of iterations of iterative refinement performed: ', solver.get_infog(15)
    )
    print('scaled residual: ', solver.get_rinfog(6))


if __name__ == '__main__':
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.mpc.interfaces.copy_values import copy_values_at_time


class DynamicVarLinker:
    """
    The purpose of this class is so that we do not have
    to call find_component or construct ComponentUIDs in a loop
    when transferring values between two different dynamic models.
    It also allows us to transfer values between variables that
    have different names in different models.

    """

    def __init__(
        self, source_variables, target_variables, source_time=None, target_time=None
    ):
        # Right now all the transfers I can think of only happen
        # in one direction
        if len(source_variables) != len(target_variables):
            raise ValueError(
                "%s must be provided two lists of time-indexed variables "
                "of equal length. Got lengths %s and %s"
                % (type(self), len(source_variables), len(target_variables))
            )
        self._source_variables = source_variables
        self._target_variables = target_variables
        self._source_time = source_time
        self._target_time = target_time

    def transfer(self, t_source=None, t_target=None):
        if t_source is None and self._source_time is None:
            raise RuntimeError(
                "Source time points were not provided in the transfer method "
                "or in the constructor."
            )
        elif t_source is None:
            t_source = self._source_time
        if t_target is None and self._target_time is None:
            raise RuntimeError(
                "Target time points were not provided in the transfer method "
                "or in the constructor."
            )
        elif t_target is None:
            t_target = self._target_time
        copy_values_at_time(
            self._source_variables, self._target_variables, t_source, t_target
        )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import enum
from pyomo.opt.results.container import MapContainer, ScalarType


#
# A coarse summary of how the solver terminated.
#
class SolverStatus(str, enum.Enum):
    ok = 'ok'  # Normal termination
    warning = 'warning'  # Termination with unusual condition
    error = 'error'  # Terminated internally with error
    aborted = 'aborted'  # Terminated due to external conditions
    #   (e.g. interrupts)
    unknown = 'unknown'  # An uninitialized value

    # Overloading __str__ is needed to match the behavior of the old
    # pyutilib.enum class (removed June 2020). There are spots in the
    # code base that expect the string representation for items in the
    # enum to not include the class name. New uses of enum shouldn't
    # need to do this.
    def __str__(self):
        return self.value


#
# A description of how the solver terminated
#
class TerminationCondition(str, enum.Enum):
    # UNKNOWN
    unknown = 'unknown'  # An uninitialized value
    # OK
    maxTimeLimit = 'maxTimeLimit'  # Exceeded maximum time limited allowed by user
    #    but having return a feasible solution
    maxIterations = 'maxIterations'  # Exceeded maximum number of iterations allowed
    #    by user (e.g., simplex iterations)
    minFunctionValue = (
        'minFunctionValue'  # Found solution smaller than specified function
    )
    #    value
    minStepLength = 'minStepLength'  # Step length is smaller than specified limit
    globallyOptimal = 'globallyOptimal'  # Found a globally optimal solution
    locallyOptimal = 'locallyOptimal'  # Found a locally optimal solution
    feasible = 'feasible'  # Found a solution that is feasible
    optimal = 'optimal'  # Found an optimal solution
    maxEvaluations = 'maxEvaluations'  # Exceeded maximum number of problem evaluations
    #    (e.g., branch and bound nodes)
    other = 'other'  # Other, uncategorized normal termination
    # WARNING
    unbounded = 'unbounded'  # Demonstrated that problem is unbounded
    infeasible = 'infeasible'  # Demonstrated that the problem is infeasible
    infeasibleOrUnbounded = (
        'infeasibleOrUnbounded'  # Problem is either infeasible or unbounded
    )
    invalidProblem = 'invalidProblem'  # The problem setup or characteristics are not
    #    valid for the solver
    intermediateNonInteger = (
        'intermediateNonInteger'  # A non-integer solution has been returned
    )
    noSolution = 'noSolution'  # No feasible solution found but infeasibility
    #    not proven
    # ERROR
    solverFailure = 'solverFailure'  # Solver failed to terminate correctly
    internalSolverError = 'internalSolverError'  # Internal solver error
    error = 'error'  # Other errors
    # ABORTED
    userInterrupt = 'userInterrupt'  # Interrupt signal generated by user
    resourceInterrupt = 'resourceInterrupt'  # Interrupt signal in resources used by
    #    optimizer
    licensingProblems = 'licensingProblems'  # Problem accessing solver license

    # Overloading __str__ is needed to match the behavior of the old
    # pyutilib.enum class (removed June 2020). There are spots in the
    # code base that expect the string representation for items in the
    # enum to not include the class name. New uses of enum shouldn't
    # need to do this.
    def __str__(self):
        return self.value

    @staticmethod
    def to_solver_status(tc):
        """Maps a TerminationCondition to SolverStatus based on enum value

        Parameters
        ----------
        tc: TerminationCondition

        Returns
        -------
        SolverStatus
        """
        if tc in {
            TerminationCondition.maxTimeLimit,
            TerminationCondition.maxIterations,
            TerminationCondition.minFunctionValue,
            TerminationCondition.minStepLength,
            TerminationCondition.globallyOptimal,
            TerminationCondition.locallyOptimal,
            TerminationCondition.feasible,
            TerminationCondition.optimal,
            TerminationCondition.maxEvaluations,
            TerminationCondition.other,
        }:
            return SolverStatus.ok
        if tc in {
            TerminationCondition.unbounded,
            TerminationCondition.infeasible,
            TerminationCondition.infeasibleOrUnbounded,
            TerminationCondition.invalidProblem,
            TerminationCondition.intermediateNonInteger,
            TerminationCondition.noSolution,
        }:
            return SolverStatus.warning
        if tc in {
            TerminationCondition.solverFailure,
            TerminationCondition.internalSolverError,
            TerminationCondition.error,
        }:
            return SolverStatus.error
        if tc in {
            TerminationCondition.userInterrupt,
            TerminationCondition.resourceInterrupt,
            TerminationCondition.licensingProblems,
        }:
            return SolverStatus.aborted
        return SolverStatus.unknown


def check_optimal_termination(results):
    """
    This function returns True if the termination condition for the solver
    is 'optimal'. This is backwards and forwards compatible.

    Parameters
    ----------
    results : Pyomo Results object returned from solver.solve

    Returns
    -------
    `bool`
    """
    from pyomo.contrib.solver.common.results import (
        SolutionStatus as FutureSolutionStatus,
    )

    if hasattr(results, 'solution_status'):
        if results.solution_status == FutureSolutionStatus.optimal:
            return True
    else:
        if results.solver.status == SolverStatus.ok and (
            results.solver.termination_condition
            in (
                TerminationCondition.optimal,
                TerminationCondition.locallyOptimal,
                TerminationCondition.globallyOptimal,
            )
        ):
            return True
    return False


def assert_optimal_termination(results):
    """
    This function checks if the termination condition for the solver
    is 'optimal', 'locallyOptimal', or 'globallyOptimal', and the status is 'ok'
    and it raises a RuntimeError exception if this is not true. This is backwards
    and forwards compatible.

    Parameters
    ----------
    results : Pyomo Results object returned from solver.solve
    """
    if not check_optimal_termination(results):
        if hasattr(results, 'solution_status'):
            msg = (
                'Solver failed to return an optimal solution. '
                f'Solution status: {results.solution_status}, '
                f'Termination condition: {results.termination_condition}'
            )
        else:
            msg = (
                'Solver failed to return an optimal solution. '
                f'Solution status: {results.solver.status}, '
                f'Termination condition: {results.solver.termination_condition}'
            )
        raise RuntimeError(msg)


class BranchAndBoundStats(MapContainer):
    def __init__(self):
        MapContainer.__init__(self)
        self.declare('number of bounded subproblems')
        self.declare('number of created subproblems')


class BlackBoxStats(MapContainer):
    def __init__(self):
        MapContainer.__init__(self)
        self.declare('number of function evaluations')
        self.declare('number of gradient evaluations')
        self.declare('number of iterations')


class SolverStatistics(MapContainer):
    def __init__(self):
        MapContainer.__init__(self)
        self.declare("branch_and_bound", value=BranchAndBoundStats(), active=False)
        self.declare("black_box", value=BlackBoxStats(), active=False)


class SolverInformation(MapContainer):
    def __init__(self):
        MapContainer.__init__(self)
        self.declare('name')
        self.declare('status', value=SolverStatus.ok)
        # Semantics: the integer return code from the shell in which the solver
        # is launched.
        self.declare('return_code')
        self.declare('message')
        self.declare('user_time', type=ScalarType.time)
        self.declare('system_time', type=ScalarType.time)
        self.declare('wallclock_time', type=ScalarType.time)
        # Semantics: The specific condition that caused the solver to
        # terminate.
        self.declare('termination_condition', value=TerminationCondition.unknown)
        # Semantics: A string printed by the solver that summarizes the
        # termination status.
        self.declare('termination_message')
        self.declare('statistics', value=SolverStatistics(), active=False)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import itertools
from unittest import skipUnless
import pyomo.common.unittest as unittest
from pyomo.contrib.piecewise.ordered_3d_j1_triangulation_data import (
    get_hamiltonian_paths,
    _get_double_cube_graph,
)
from pyomo.contrib.piecewise.triangulations import (
    get_unordered_j1_triangulation,
    get_ordered_j1_triangulation,
    _get_Gn_hamiltonian,
    _get_grid_hamiltonian,
)
from pyomo.common.dependencies import numpy as np, numpy_available, networkx_available
from math import factorial
import itertools


class TestTriangulations(unittest.TestCase):

    # check basic functionality for the unordered j1 triangulation.
    @unittest.skipUnless(numpy_available, "numpy is not available")
    def test_J1_small(self):
        points = [
            [0.5, 0.5],  # 0
            [0.5, 1.5],  # 1
            [0.5, 2.5],  # 2
            [1.5, 0.5],  # 3
            [1.5, 1.5],  # 4
            [1.5, 2.5],  # 5
            [2.5, 0.5],  # 6
            [2.5, 1.5],  # 7
            [2.5, 2.5],  # 8
        ]
        triangulation = get_unordered_j1_triangulation(points, 2)
        self.assertTrue(
            np.array_equal(
                triangulation.simplices,
                np.array(
                    [
                        [0, 1, 4],
                        [1, 2, 4],
                        [4, 6, 7],
                        [4, 7, 8],
                        [0, 3, 4],
                        [2, 4, 5],
                        [3, 4, 6],
                        [4, 5, 8],
                    ]
                ),
            )
        )

    def check_J1_ordered(self, points, num_points, dim):
        ordered_triangulation = get_ordered_j1_triangulation(points, dim).simplices
        self.assertEqual(
            len(ordered_triangulation), factorial(dim) * (num_points - 1) ** dim
        )
        for idx, first_simplex in enumerate(ordered_triangulation):
            if idx != len(ordered_triangulation) - 1:
                second_simplex = ordered_triangulation[idx + 1]
                # test property (2) which also guarantees property (1) (from Vielma 2010)
                self.assertEqual(
                    first_simplex[-1],
                    second_simplex[0],
                    msg="Last and first vertices of adjacent simplices did not match",
                )
                # The way I am constructing these, they should always share an (n-1)-face.
                # Check that too for good measure.
                count = len(set(first_simplex).intersection(set(second_simplex)))
                self.assertEqual(count, dim)  # (n-1)-face has n points

    @unittest.skipUnless(numpy_available, "numpy is not available")
    def test_J1_ordered_2d(self):
        self.check_J1_ordered(list(itertools.product([0, 1, 2], [1, 2.4, 3])), 3, 2)
        self.check_J1_ordered(
            list(itertools.product([0, 1, 2, 4, 5], [1, 2.4, 3, 5, 6])), 5, 2
        )
        self.check_J1_ordered(
            list(
                itertools.product([0, 1, 2, 4, 5, 6.3, 7.1], [1, 2.4, 3, 5, 6, 9.1, 10])
            ),
            7,
            2,
        )
        self.check_J1_ordered(
            list(
                itertools.product(
                    [0, 1, 2, 4, 5, 6.3, 7.1, 7.2, 7.3],
                    [1, 2.4, 3, 5, 6, 9.1, 10, 11, 12],
                )
            ),
            9,
            2,
        )

    @unittest.skipUnless(numpy_available, "numpy is not available")
    def test_J1_ordered_3d(self):
        self.check_J1_ordered(
            list(itertools.product([0, 1, 2], [1, 2.4, 3], [2, 3, 4])), 3, 3
        )
        self.check_J1_ordered(
            list(
                itertools.product([0, 1, 2, 4, 5], [1, 2.4, 3, 5, 6], [-1, 0, 1, 2, 3])
            ),
            5,
            3,
        )
        self.check_J1_ordered(
            list(
                itertools.product(
                    [0, 1, 2, 4, 5, 6, 7],
                    [1, 2.4, 3, 5, 6, 6.5, 7],
                    [-1, 0, 1, 2, 3, 4, 5],
                )
            ),
            7,
            3,
        )
        self.check_J1_ordered(
            list(
                itertools.product(
                    [0, 1, 2, 4, 5, 6, 7, 8, 9],
                    [1, 2.4, 3, 5, 6, 6.5, 7, 8, 9],
                    [-1, 0, 1, 2, 3, 4, 5, 6, 7],
                )
            ),
            9,
            3,
        )

    @unittest.skipUnless(numpy_available, "numpy is not available")
    def test_J1_ordered_4d_and_above(self):
        self.check_J1_ordered(
            list(
                itertools.product(
                    [0, 1, 2, 4, 5],
                    [1, 2.4, 3, 5, 6],
                    [-1, 0, 1, 2, 3],
                    [1, 2, 3, 4, 5],
                )
            ),
            5,
            4,
        )
        self.check_J1_ordered(
            list(
                itertools.product(
                    [0, 1, 2, 4, 5],
                    [1, 2.4, 3, 5, 6],
                    [-1, 0, 1, 2, 3],
                    [1, 2, 3, 4, 5],
                    [2, 3, 4, 5, 6],
                )
            ),
            5,
            5,
        )

    def check_Gn_hamiltonian_path(self, n, start_permutation, target_symbol, last):
        path = _get_Gn_hamiltonian(n, start_permutation, target_symbol, last)
        self.assertEqual(len(path), factorial(n))
        self.assertEqual(path[0], start_permutation)
        if last:
            self.assertEqual(path[-1][-1], target_symbol)
        else:
            self.assertEqual(path[-1][0], target_symbol)
        for pi in itertools.permutations(range(1, n + 1), n):
            self.assertTrue(tuple(pi) in path)
        for i in range(len(path) - 1):
            diff_indices = [j for j in range(n) if path[i][j] != path[i + 1][j]]
            self.assertEqual(len(diff_indices), 2)
            self.assertEqual(diff_indices[0], diff_indices[1] - 1)
            self.assertEqual(path[i][diff_indices[0]], path[i + 1][diff_indices[1]])
            self.assertEqual(path[i][diff_indices[1]], path[i + 1][diff_indices[0]])

    def test_Gn_hamiltonian_paths(self):
        # each of the base cases
        self.check_Gn_hamiltonian_path(4, (1, 2, 3, 4), 1, False)
        self.check_Gn_hamiltonian_path(4, (1, 2, 3, 4), 2, False)
        self.check_Gn_hamiltonian_path(4, (1, 2, 3, 4), 3, False)
        self.check_Gn_hamiltonian_path(4, (1, 2, 3, 4), 4, False)
        # some variants with start permutations and/or last
        self.check_Gn_hamiltonian_path(4, (3, 4, 1, 2), 2, False)
        self.check_Gn_hamiltonian_path(4, (1, 3, 2, 4), 3, True)
        self.check_Gn_hamiltonian_path(4, (1, 4, 2, 3), 4, True)
        self.check_Gn_hamiltonian_path(4, (1, 2, 3, 4), 2, True)
        # some recursive cases
        self.check_Gn_hamiltonian_path(5, (1, 2, 3, 4, 5), 1, False)
        self.check_Gn_hamiltonian_path(5, (1, 2, 3, 4, 5), 3, False)
        self.check_Gn_hamiltonian_path(5, (1, 2, 3, 4, 5), 5, False)
        self.check_Gn_hamiltonian_path(5, (1, 2, 4, 3, 5), 5, True)
        self.check_Gn_hamiltonian_path(6, (6, 1, 2, 4, 3, 5), 5, True)
        self.check_Gn_hamiltonian_path(6, (6, 1, 2, 4, 3, 5), 5, False)
        self.check_Gn_hamiltonian_path(7, (1, 2, 3, 4, 5, 6, 7), 7, False)

    def check_grid_hamiltonian(self, dim, length):
        path = _get_grid_hamiltonian(dim, length)
        self.assertEqual(len(path), length**dim)
        for x in itertools.product(range(length), repeat=dim):
            self.assertTrue(list(x) in path)
        for i in range(len(path) - 1):
            diff_indices = [j for j in range(dim) if path[i][j] != path[i + 1][j]]
            self.assertEqual(len(diff_indices), 1)
            self.assertEqual(
                abs(path[i][diff_indices[0]] - path[i + 1][diff_indices[0]]), 1
            )

    def test_grid_hamiltonian_paths(self):
        self.check_grid_hamiltonian(1, 5)
        self.check_grid_hamiltonian(2, 5)
        self.check_grid_hamiltonian(2, 8)
        self.check_grid_hamiltonian(3, 5)
        self.check_grid_hamiltonian(4, 3)


@unittest.skipUnless(networkx_available, "Networkx is not available")
class TestHamiltonianPaths(unittest.TestCase):
    def test_hamiltonian_paths(self):
        G = _get_double_cube_graph()

        paths = get_hamiltonian_paths()
        self.assertEqual(len(paths), 60)

        for ((s1, t1), (s2, t2)), path in paths.items():
            # ESJ: I'm not quite sure how to check this is *the right* path
            # given the key?

            # Check it's Hamiltonian
            self.assertEqual(len(path), 48)
            # Check it's a path
            for idx in range(1, 48):
                self.assertTrue(G.has_edge(path[idx - 1], path[idx]))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from .base import (
    DirectLinearSolverInterface,
    LinearSolverStatus,
    LinearSolverResults,
    LinearSolverInterface,
)
from scipy.sparse.linalg import splu, LinearOperator
from scipy.linalg import eigvals
from scipy.sparse import isspmatrix_csc, spmatrix
from pyomo.contrib.pynumero.sparse import BlockVector, BlockMatrix
import numpy as np
from typing import Union, Tuple, Optional, Callable


class ScipyLU(DirectLinearSolverInterface):
    def __init__(self):
        self._lu = None

    def do_symbolic_factorization(
        self, matrix: Union[spmatrix, BlockMatrix], raise_on_error: bool = True
    ) -> LinearSolverResults:
        res = LinearSolverResults()
        res.status = LinearSolverStatus.successful
        return res

    def do_numeric_factorization(
        self, matrix: Union[spmatrix, BlockMatrix], raise_on_error: bool = True
    ) -> LinearSolverResults:
        if not isspmatrix_csc(matrix):
            matrix = matrix.tocsc()
        res = LinearSolverResults()
        try:
            self._lu = splu(matrix)
            res.status = LinearSolverStatus.successful
        except RuntimeError as err:
            if raise_on_error:
                raise err
            if "Factor is exactly singular" in str(err):
                res.status = LinearSolverStatus.singular
            else:
                res.status = LinearSolverStatus.error

        return res

    def do_back_solve(
        self, rhs: Union[np.ndarray, BlockVector], raise_on_error: bool = True
    ) -> Tuple[Optional[Union[np.ndarray, BlockVector]], LinearSolverResults]:
        if isinstance(rhs, BlockVector):
            _rhs = rhs.flatten()
        else:
            _rhs = rhs

        result = self._lu.solve(_rhs)

        if isinstance(rhs, BlockVector):
            _result = rhs.copy_structure()
            _result.copyfrom(result)
            result = _result

        return result, LinearSolverResults(LinearSolverStatus.successful)


class _LinearOperator(LinearOperator):
    def __init__(self, matrix: Union[spmatrix, BlockMatrix]):
        self._matrix = matrix
        shape = self._matrix.shape
        dtype = self._matrix.dtype
        super(_LinearOperator, self).__init__(shape=shape, dtype=dtype)

    def _matvec(self, x):
        return self._matrix * x

    def _adjoint(self):
        return _LinearOperator(self._matrix.transpose())


class ScipyIterative(LinearSolverInterface):
    def __init__(self, method: Callable, options=None):
        self.method = method
        if options is None:
            self.options = dict()
        else:
            self.options = dict(options)

    def solve(
        self,
        matrix: Union[spmatrix, BlockMatrix],
        rhs: Union[np.ndarray, BlockVector],
        raise_on_error: bool = True,
    ) -> Tuple[Optional[Union[np.ndarray, BlockVector]], LinearSolverResults]:
        # eventually, we will want to remove the .tocoo(), but we would have to first
        # figure out how to deal with np.asarray for BlockVector and MPIBlockVector
        # np.asarray is used on rhs within the scipy iterative solvers
        linear_operator = _LinearOperator(matrix.tocoo())
        if isinstance(rhs, BlockVector):
            _rhs = rhs.flatten()
        else:
            _rhs = rhs
        result, info = self.method(linear_operator, _rhs, **self.options)
        if info == 0:
            stat = LinearSolverStatus.successful
        else:
            stat = LinearSolverStatus.error

        if isinstance(rhs, BlockVector):
            _result = rhs.copy_structure()
            _result.copyfrom(result)
            result = _result

        return result, LinearSolverResults(stat)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import random
from pyomo.contrib.incidence_analysis.matching import maximum_matching
from pyomo.contrib.incidence_analysis.triangularize import (
    get_scc_of_projection,
    block_triangularize,
    map_coords_to_block_triangular_indices,
    get_diagonal_blocks,
)
from pyomo.common.dependencies import (
    scipy,
    scipy_available,
    networkx as nx,
    networkx_available,
)

if scipy_available:
    sps = scipy.sparse
if networkx_available:
    nxb = nx.algorithms.bipartite

import pyomo.common.unittest as unittest


@unittest.skipUnless(networkx_available, "networkx is not available")
@unittest.skipUnless(scipy_available, "scipy is not available")
class TestGetSCCOfProjection(unittest.TestCase):
    def test_graph_decomposable_tridiagonal_shuffled(self):
        """
        This is the same graph as in test_decomposable_tridiagonal_shuffled
        below, but now we convert the matrix into a bipartite graph and
        use get_scc_of_projection.

        The matrix decomposes into 2x2 blocks:
        |x x      |
        |x x      |
        |  x x x  |
        |    x x  |
        |      x x|
        """
        N = 11
        row = []
        col = []
        data = []

        # Diagonal
        row.extend(range(N))
        col.extend(range(N))
        data.extend(1 for _ in range(N))

        # Below diagonal
        row.extend(range(1, N))
        col.extend(range(N - 1))
        data.extend(1 for _ in range(N - 1))

        # Above diagonal
        row.extend(i for i in range(N - 1) if not i % 2)
        col.extend(i + 1 for i in range(N - 1) if not i % 2)
        data.extend(1 for i in range(N - 1) if not i % 2)

        # Same results hold after applying a random permutation.
        row_perm = list(range(N))
        col_perm = list(range(N))
        random.shuffle(row_perm)
        random.shuffle(col_perm)

        row = [row_perm[i] for i in row]
        col = [col_perm[j] for j in col]

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))
        graph = nxb.matrix.from_biadjacency_matrix(matrix)
        row_nodes = list(range(N))
        sccs = get_scc_of_projection(graph, row_nodes)

        self.assertEqual(len(sccs), (N + 1) // 2)

        for i in range((N + 1) // 2):
            # Note that these rows and cols are in the permuted space
            rows = set(r for r, _ in sccs[i])
            cols = set(c - N for _, c in sccs[i])

            pred_rows = {row_perm[2 * i]}
            pred_cols = {col_perm[2 * i]}

            if 2 * i + 1 < N:
                pred_rows.add(row_perm[2 * i + 1])
                pred_cols.add(col_perm[2 * i + 1])

            self.assertEqual(pred_rows, rows)
            self.assertEqual(pred_cols, cols)

    def test_scc_exceptions(self):
        graph = nx.Graph()
        graph.add_nodes_from(range(3))
        graph.add_edges_from([(0, 1), (0, 2), (1, 2)])
        top_nodes = [0]
        msg = "graph is not bipartite"
        with self.assertRaisesRegex(RuntimeError, msg):
            sccs = get_scc_of_projection(graph, top_nodes=top_nodes)

        graph = nx.Graph()
        graph.add_nodes_from(range(3))
        graph.add_edges_from([(0, 1), (0, 2)])
        top_nodes[0]
        msg = "bipartite sets of different cardinalities"
        with self.assertRaisesRegex(RuntimeError, msg):
            sccs = get_scc_of_projection(graph, top_nodes=top_nodes)

        graph = nx.Graph()
        graph.add_nodes_from(range(4))
        graph.add_edges_from([(0, 1), (0, 2)])
        top_nodes = [0, 3]
        msg = "without a perfect matching"
        with self.assertRaisesRegex(RuntimeError, msg):
            sccs = get_scc_of_projection(graph, top_nodes=top_nodes)


@unittest.skipUnless(networkx_available, "networkx is not available")
@unittest.skipUnless(scipy_available, "scipy is not available")
class TestTriangularize(unittest.TestCase):
    def test_low_rank_exception(self):
        N = 5
        row = list(range(N - 1))
        col = list(range(N - 1))
        data = [1 for _ in range(N - 1)]

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))

        with self.assertRaises(RuntimeError) as exc:
            row_block_map, col_block_map = block_triangularize(matrix)
        self.assertIn('perfect matching', str(exc.exception))

    def test_non_square_exception(self):
        N = 5
        row = list(range(N - 1))
        col = list(range(N - 1))
        data = [1 for _ in range(N - 1)]

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N - 1))

        with self.assertRaises(ValueError) as exc:
            row_block_map, col_block_map = block_triangularize(matrix)
        self.assertIn('non-square matrices', str(exc.exception))

    def test_identity(self):
        N = 5
        matrix = sps.identity(N).tocoo()
        row_block_map, col_block_map = map_coords_to_block_triangular_indices(matrix)
        row_values = set(row_block_map.values())
        col_values = set(row_block_map.values())

        # For a (block) diagonal matrix, the order of diagonal
        # blocks is arbitrary, so we can't perform any strong
        # checks here.
        #
        # Perfect matching is unique, but order of strongly
        # connected components is not.

        self.assertEqual(len(row_block_map), N)
        self.assertEqual(len(col_block_map), N)
        self.assertEqual(len(row_values), N)
        self.assertEqual(len(col_values), N)

        for i in range(N):
            self.assertIn(i, row_block_map)
            self.assertIn(i, col_block_map)
            self.assertIn(i, row_values)
            self.assertIn(i, col_values)

    def test_lower_tri(self):
        """
        This matrix has a unique maximal matching and SCC
        order, making it a good test for a "fully decomposable"
        matrix.
        |x        |
        |x x      |
        |  x x    |
        |    x x  |
        |      x x|
        """
        N = 5
        row = []
        col = []
        data = []
        # Diagonal
        row.extend(range(N))
        col.extend(range(N))
        data.extend(1 for _ in range(N))

        # Below diagonal
        row.extend(range(1, N))
        col.extend(range(N - 1))
        data.extend(1 for _ in range(N - 1))

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))

        row_block_map, col_block_map = map_coords_to_block_triangular_indices(matrix)
        row_values = set(row_block_map.values())
        col_values = set(row_block_map.values())

        self.assertEqual(len(row_values), N)
        self.assertEqual(len(col_values), N)

        for i in range(N):
            self.assertEqual(row_block_map[i], i)
            self.assertEqual(col_block_map[i], i)

    def test_upper_tri(self):
        """
        This matrix has a unique maximal matching and SCC
        order, making it a good test for a "fully decomposable"
        matrix.
        |x x      |
        |  x x    |
        |    x x  |
        |      x x|
        |        x|
        """
        N = 5
        row = []
        col = []
        data = []
        # Diagonal
        row.extend(range(N))
        col.extend(range(N))
        data.extend(1 for _ in range(N))

        # Below diagonal
        row.extend(range(N - 1))
        col.extend(range(1, N))
        data.extend(1 for _ in range(N - 1))

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))

        row_block_map, col_block_map = map_coords_to_block_triangular_indices(matrix)
        row_values = set(row_block_map.values())
        col_values = set(row_block_map.values())

        self.assertEqual(len(row_values), N)
        self.assertEqual(len(col_values), N)

        for i in range(N):
            # The block_triangularize function permutes
            # to lower triangular form, so rows and
            # columns are transposed to assemble the blocks.
            self.assertEqual(row_block_map[i], N - 1 - i)
            self.assertEqual(col_block_map[i], N - 1 - i)

    def test_bordered(self):
        """
        This matrix is non-decomposable
        |x       x|
        |  x     x|
        |    x   x|
        |      x x|
        |x x x x  |
        """
        N = 5
        row = []
        col = []
        data = []
        # Diagonal
        row.extend(range(N - 1))
        col.extend(range(N - 1))
        data.extend(1 for _ in range(N - 1))

        # Bottom row
        row.extend(N - 1 for _ in range(N - 1))
        col.extend(range(N - 1))
        data.extend(1 for _ in range(N - 1))

        # Right column
        row.extend(range(N - 1))
        col.extend(N - 1 for _ in range(N - 1))
        data.extend(1 for _ in range(N - 1))

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))

        row_block_map, col_block_map = map_coords_to_block_triangular_indices(matrix)
        row_values = set(row_block_map.values())
        col_values = set(row_block_map.values())

        self.assertEqual(len(row_values), 1)
        self.assertEqual(len(col_values), 1)

        for i in range(N):
            self.assertEqual(row_block_map[i], 0)
            self.assertEqual(col_block_map[i], 0)

    def test_decomposable_bordered(self):
        """
        This matrix decomposes
        |x        |
        |  x      |
        |    x   x|
        |      x x|
        |x x x x  |
        """
        N = 5
        half = N // 2
        row = []
        col = []
        data = []

        # Diagonal
        row.extend(range(N - 1))
        col.extend(range(N - 1))
        data.extend(1 for _ in range(N - 1))

        # Bottom row
        row.extend(N - 1 for _ in range(N - 1))
        col.extend(range(N - 1))
        data.extend(1 for _ in range(N - 1))

        # Right column
        row.extend(range(half, N - 1))
        col.extend(N - 1 for _ in range(half, N - 1))
        data.extend(1 for _ in range(half, N - 1))

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))

        row_block_map, col_block_map = map_coords_to_block_triangular_indices(matrix)
        row_values = set(row_block_map.values())
        col_values = set(row_block_map.values())

        self.assertEqual(len(row_values), half + 1)
        self.assertEqual(len(col_values), half + 1)

        first_half_set = set(range(half))
        for i in range(N):
            if i < half:
                # The first N//2 diagonal blocks are unordered
                self.assertIn(row_block_map[i], first_half_set)
                self.assertIn(col_block_map[i], first_half_set)
            else:
                self.assertEqual(row_block_map[i], half)
                self.assertEqual(col_block_map[i], half)

    def test_decomposable_tridiagonal(self):
        """
        This matrix decomposes into 2x2 blocks
        |x x      |
        |x x      |
        |  x x x  |
        |    x x  |
        |      x x|
        """
        N = 5
        row = []
        col = []
        data = []

        # Diagonal
        row.extend(range(N))
        col.extend(range(N))
        data.extend(1 for _ in range(N))

        # Below diagonal
        row.extend(range(1, N))
        col.extend(range(N - 1))
        data.extend(1 for _ in range(N - 1))

        # Above diagonal
        row.extend(i for i in range(N - 1) if not i % 2)
        col.extend(i + 1 for i in range(N - 1) if not i % 2)
        data.extend(1 for i in range(N - 1) if not i % 2)

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))

        row_block_map, col_block_map = map_coords_to_block_triangular_indices(matrix)
        row_values = set(row_block_map.values())
        col_values = set(row_block_map.values())

        self.assertEqual(len(row_values), (N + 1) // 2)
        self.assertEqual(len(col_values), (N + 1) // 2)

        for i in range((N + 1) // 2):
            self.assertEqual(row_block_map[2 * i], i)
            self.assertEqual(col_block_map[2 * i], i)

            if 2 * i + 1 < N:
                self.assertEqual(row_block_map[2 * i + 1], i)
                self.assertEqual(col_block_map[2 * i + 1], i)

    def test_decomposable_tridiagonal_shuffled(self):
        """
        This matrix decomposes into 2x2 blocks
        |x x      |
        |x x      |
        |  x x x  |
        |    x x  |
        |      x x|
        """
        N = 5
        row = []
        col = []
        data = []

        # Diagonal
        row.extend(range(N))
        col.extend(range(N))
        data.extend(1 for _ in range(N))

        # Below diagonal
        row.extend(range(1, N))
        col.extend(range(N - 1))
        data.extend(1 for _ in range(N - 1))

        # Above diagonal
        row.extend(i for i in range(N - 1) if not i % 2)
        col.extend(i + 1 for i in range(N - 1) if not i % 2)
        data.extend(1 for i in range(N - 1) if not i % 2)

        # Same results hold after applying a random permutation.
        row_perm = list(range(N))
        col_perm = list(range(N))
        random.shuffle(row_perm)
        random.shuffle(col_perm)

        row = [row_perm[i] for i in row]
        col = [col_perm[j] for j in col]

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))

        row_block_map, col_block_map = map_coords_to_block_triangular_indices(matrix)
        row_values = set(row_block_map.values())
        col_values = set(row_block_map.values())

        self.assertEqual(len(row_values), (N + 1) // 2)
        self.assertEqual(len(col_values), (N + 1) // 2)

        for i in range((N + 1) // 2):
            row_idx = row_perm[2 * i]
            col_idx = col_perm[2 * i]
            self.assertEqual(row_block_map[row_idx], i)
            self.assertEqual(col_block_map[col_idx], i)

            if 2 * i + 1 < N:
                row_idx = row_perm[2 * i + 1]
                col_idx = col_perm[2 * i + 1]
                self.assertEqual(row_block_map[row_idx], i)
                self.assertEqual(col_block_map[col_idx], i)

    def test_decomposable_tridiagonal_diagonal_blocks(self):
        """
        This matrix decomposes into 2x2 blocks
        |x x      |
        |x x      |
        |  x x x  |
        |    x x  |
        |      x x|
        """
        N = 5
        row = []
        col = []
        data = []

        # Diagonal
        row.extend(range(N))
        col.extend(range(N))
        data.extend(1 for _ in range(N))

        # Below diagonal
        row.extend(range(1, N))
        col.extend(range(N - 1))
        data.extend(1 for _ in range(N - 1))

        # Above diagonal
        row.extend(i for i in range(N - 1) if not i % 2)
        col.extend(i + 1 for i in range(N - 1) if not i % 2)
        data.extend(1 for i in range(N - 1) if not i % 2)

        matrix = sps.coo_matrix((data, (row, col)), shape=(N, N))

        row_blocks, col_blocks = get_diagonal_blocks(matrix)

        self.assertEqual(len(row_blocks), (N + 1) // 2)
        self.assertEqual(len(col_blocks), (N + 1) // 2)

        for i in range((N + 1) // 2):
            rows = row_blocks[i]
            cols = col_blocks[i]

            if 2 * i + 1 < N:
                self.assertEqual(set(rows), {2 * i, 2 * i + 1})
                self.assertEqual(set(cols), {2 * i, 2 * i + 1})
            else:
                self.assertEqual(set(rows), {2 * i})
                self.assertEqual(set(cols), {2 * i})


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import inspect
import importlib
import importlib.util
import logging
import sys
import warnings

from collections.abc import Mapping
from types import ModuleType
from typing import List

from pyomo.common.deprecation import deprecated, deprecation_warning
from pyomo.common.errors import DeferredImportError
from pyomo.common.flags import (
    in_testing_environment,
    building_documentation,
    serializing,
)

SUPPRESS_DEPENDENCY_WARNINGS = False


class ModuleUnavailable:
    """Mock object that raises :py:class:`.DeferredImportError` upon attribute access

    This object is returned by :py:func:`attempt_import()` in lieu of
    the module in the case that the module import fails.  Any attempts
    to access attributes on this object will raise a :py:class:`.DeferredImportError`
    exception.

    Parameters
    ----------
    name: str
        The module name that was being imported

    message: str
        The string message to return in the raised exception

    version_error: str
        A string to add to the message if the module failed to import because
        it did not match the required version

    import_error: str
        A string to add to the message documenting the Exception
        raised when the module failed to import.

    package: str
        The module name that originally attempted the import
    """

    _getattr_raises_attributeerror = {
        # We need special handling for Sphinx here, as it will look for the
        # __sphinx_mock__ attribute on all module-level objects, and we need
        # that to raise an AttributeError and not a DeferredImportError
        '__sphinx_mock__',
        # We need special handling for dill as well, as dill attempts to
        # pickle module globals by looking for the '_dill' attribute on
        # all global objects.
        '_dill',
    }

    def __init__(self, name, message, version_error, import_error, package):
        self.__name__ = name
        self._moduleunavailable_info_ = (message, version_error, import_error, package)

    def __getattr__(self, attr):
        if serializing() or attr in ModuleUnavailable._getattr_raises_attributeerror:
            msg = "'%s' object has no attribute '%s'" % (type(self).__name__, attr)
            raise AttributeError(msg)
        raise DeferredImportError(self._moduleunavailable_message())

    def __getstate__(self):
        return (self.__name__, self._moduleunavailable_info_)

    def __setstate__(self, state):
        self.__name__, self._moduleunavailable_info_ = state

    # Included because recent dill picklers look for the mro() when
    # detecting numpy types
    def mro(self):
        return [ModuleUnavailable, object]

    def _moduleunavailable_message(self, msg=None):
        _err, _ver, _imp, _package = self._moduleunavailable_info_
        if msg is None:
            msg = _err
        if _imp:
            if not msg or not str(msg):
                _pkg_str = _package.split('.')[0].capitalize()
                if _pkg_str:
                    _pkg_str += ' '
                msg = (
                    "The %s module (an optional %sdependency) "
                    "failed to import: %s" % (self.__name__, _pkg_str, _imp)
                )
            else:
                msg = "%s (import raised %s)" % (msg, _imp)
        if _ver:
            if not msg or not str(msg):
                msg = "The %s module %s" % (self.__name__, _ver)
            else:
                msg = "%s (%s)" % (msg, _ver)
        return msg

    def log_import_warning(self, logger='pyomo', msg=None):
        """Log the import error message to the specified logger

        This will log the the import error message to the specified
        logger.  If ``msg=`` is specified, it will override the default
        message passed to this instance of
        :py:class:`ModuleUnavailable`.

        """
        logging.getLogger(logger).warning(self._moduleunavailable_message(msg))

    @deprecated("use :py:class:`log_import_warning()`", version='6.0')
    def generate_import_warning(self, logger='pyomo.common'):
        self.log_import_warning(logger)


class DeferredImportModule:
    """Mock module object to support the deferred import of a module.

    This object is returned by :py:func:`attempt_import()` in lieu of
    the module when :py:func:`attempt_import()` is called with
    ``defer_import=True``.  Any attempts to access attributes on this
    object will trigger the actual module import and return either the
    appropriate module attribute or else if the module import fails,
    raise a :py:class:`.DeferredImportError` exception.

    """

    def __init__(self, indicator, deferred_submodules, submodule_name):
        self._indicator_flag = indicator
        self._submodule_name = submodule_name
        self.__file__ = None  # Disable coverage of this module
        self.__spec__ = None  # Indicate that this is not a "real" module

        if not deferred_submodules:
            return
        if submodule_name is None:
            submodule_name = ''
        for name in deferred_submodules:
            if not name.startswith(submodule_name + '.'):
                continue
            _local_name = name[(1 + len(submodule_name)) :]
            if '.' in _local_name:
                continue
            setattr(
                self,
                _local_name,
                DeferredImportModule(
                    indicator, deferred_submodules, submodule_name + '.' + _local_name
                ),
            )

    def __getattr__(self, attr):
        self._indicator_flag.resolve()
        _mod = self._indicator_flag._module
        if self._submodule_name:
            for _sub in self._submodule_name[1:].split('.'):
                _mod = getattr(_mod, _sub)
        return getattr(_mod, attr)

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        for k, v in state.items():
            super().__setattr__(k, v)

    # Included because recent dill picklers look for the mro() when
    # detecting numpy types
    def mro(self):
        return [DeferredImportModule, object]


def UnavailableClass(unavailable_module):
    """Function to generate an "unavailable" base class

    This function returns a custom class that wraps the
    :py:class:`ModuleUnavailable` instance returned by
    :py:func:`attempt_import` when the target module is not available.
    Any attempt to instantiate this class (or a class derived from it)
    or access a class attribute will raise the
    :py:class:`.DeferredImportError` from the wrapped
    :py:class:`ModuleUnavailable` object.

    Parameters
    ----------
    unavailable_module: ModuleUnavailable
        The :py:class:`ModuleUnavailable` instance (from
        :py:func:`attempt_import`) to use to generate the
        :py:class:`.DeferredImportError`.

    Example
    -------

    Declaring a class that inherits from an optional dependency:

    .. doctest::

       >>> from pyomo.common.dependencies import attempt_import, UnavailableClass
       >>> bogus, bogus_available = attempt_import('bogus_unavailable_class')
       >>> class MyPlugin(bogus.plugin if bogus_available else UnavailableClass(bogus)):
       ...     pass

    Attempting to instantiate the derived class generates an exception
    when the module is unavailable:

    .. doctest::

       >>> MyPlugin()
       Traceback (most recent call last):
          ...
       pyomo.common.dependencies.DeferredImportError: The class 'MyPlugin' cannot be
       created because a needed optional dependency was not found (import raised
       ModuleNotFoundError: No module named 'bogus_unavailable_class')

    As does attempting to access class attributes on the derived class:

    .. testcode::
       :hide:

       # We suppress this exception when building the documentation
       # from pyomo.common.flags import building_documentation
       building_documentation(False)

    .. doctest::

       >>> MyPlugin.create_instance()
       Traceback (most recent call last):
          ...
       pyomo.common.dependencies.DeferredImportError: The class attribute
       'MyPlugin.create_instance' is not available because a needed optional
       dependency was not found (import raised ModuleNotFoundError: No module
       named 'bogus_unavailable_class')

    .. testcode::
       :hide:

       building_documentation(None)

    """

    class UnavailableMeta(type):
        def __getattr__(cls, name):
            if building_documentation():
                # If we are building documentation, avoid the
                # DeferredImportError (we will still raise one if
                # someone attempts to *create* an instance of this
                # class)
                return getattr(super(), name)
            raise DeferredImportError(
                unavailable_module._moduleunavailable_message(
                    f"The class attribute '{cls.__name__}.{name}' is not available "
                    "because a needed optional dependency was not found"
                )
            )

    class UnavailableBase(metaclass=UnavailableMeta):
        def __new__(cls, *args, **kwargs):
            raise DeferredImportError(
                unavailable_module._moduleunavailable_message(
                    f"The class '{cls.__name__}' cannot be created because a "
                    "needed optional dependency was not found"
                )
            )

    return UnavailableBase


class _DeferredImportIndicatorBase:
    def __and__(self, other):
        return _DeferredAnd(self, other)

    def __or__(self, other):
        return _DeferredOr(self, other)

    def __rand__(self, other):
        return _DeferredAnd(other, self)

    def __ror__(self, other):
        return _DeferredOr(other, self)


class DeferredImportIndicator(_DeferredImportIndicatorBase):
    """Placeholder indicating if an import was successful.

    This object serves as a placeholder for the Boolean indicator if a
    deferred module import was successful.  Casting this instance to
    `bool` will cause the import to be attempted.  The actual import logic
    is here and not in the :py:class:`DeferredImportModule` to reduce the number of
    attributes on the :py:class:`DeferredImportModule`.

    :py:class:`DeferredImportIndicator` supports limited logical expressions
    using the ``&`` (and) and ``|`` (or) binary operators.  Creating
    these expressions does not trigger the import of the corresponding
    :py:class:`DeferredImportModule` instances, although casting the
    resulting expression to ``bool()`` will trigger any relevant
    imports.

    """

    def __init__(
        self,
        name,
        error_message,
        catch_exceptions,
        minimum_version,
        original_globals,
        callback,
        importer,
        deferred_submodules,
    ):
        self._names = [name]
        for _n in tuple(self._names):
            if '.' in _n:
                self._names.append(_n.split('.')[-1])
        self._error_message = error_message
        self._catch_exceptions = catch_exceptions
        self._minimum_version = minimum_version
        self._original_globals = original_globals
        self._callback = callback
        self._importer = importer
        self._module = None
        self._available = None
        self._deferred_submodules = deferred_submodules
        # If this import has a callback, then record this deferred
        # import so that any direct imports of this module also trigger
        # the resolution of this DeferredImportIndicator (and the
        # corresponding callback)
        if callback is not None:
            DeferredImportCallbackFinder._callbacks.setdefault(name, []).append(self)

    def __bool__(self):
        self.resolve()
        # resolve() guarantees that _available has been resolved to a bool
        assert self._available.__class__ is bool
        # The following cast is to keep static code analysis linters happy
        return bool(self._available)

    def resolve(self):
        # Only attempt the import once, then cache some form of result
        if self._module is None:
            package = self._original_globals.get('__name__', '')
            try:
                self._module, self._available = _perform_import(
                    name=self._names[0],
                    error_message=self._error_message,
                    minimum_version=self._minimum_version,
                    callback=self._callback,
                    importer=self._importer,
                    catch_exceptions=self._catch_exceptions,
                    package=package,
                )
            except Exception as e:
                # make sure that we cache the result
                self._module = ModuleUnavailable(
                    self._names[0],
                    "Exception raised when importing %s" % (self._names[0],),
                    None,
                    "%s: %s" % (type(e).__name__, e),
                    package,
                )
                self._available = False
                raise

            # If this module was not found, then we need to check for
            # deferred submodules and resolve them as well
            if self._deferred_submodules and type(self._module) is ModuleUnavailable:
                info = self._module._moduleunavailable_info_
                for submod in self._deferred_submodules:
                    refmod = self._module
                    for name in submod.split('.')[1:]:
                        try:
                            refmod = getattr(refmod, name)
                        except DeferredImportError:
                            setattr(
                                refmod,
                                name,
                                ModuleUnavailable(refmod.__name__ + submod, *info),
                            )
                            refmod = getattr(refmod, name)

            # Replace myself in the original globals() where I was
            # declared
            self.replace_self_in_globals(self._original_globals)

        # Replace myself in the caller globals (to avoid calls to
        # this method in the future)
        _frame = inspect.currentframe().f_back
        while _frame.f_globals is globals():
            _frame = _frame.f_back
        self.replace_self_in_globals(_frame.f_globals)

    def replace_self_in_globals(self, _globals):
        for k, v in _globals.items():
            if v is self:
                _globals[k] = self._available
            elif v.__class__ is DeferredImportModule and v._indicator_flag is self:
                if v._submodule_name is None:
                    _globals[k] = self._module
                else:
                    _mod_path = v._submodule_name.split('.')[1:]
                    _mod = self._module
                    for _sub in _mod_path:
                        _mod = getattr(_mod, _sub)
                    _globals[k] = _mod


class _DeferredAnd(_DeferredImportIndicatorBase):
    def __init__(self, a, b):
        self._a = a
        self._b = b

    def __bool__(self):
        return bool(self._a) and bool(self._b)


class _DeferredOr(_DeferredImportIndicatorBase):
    def __init__(self, a, b):
        self._a = a
        self._b = b

    def __bool__(self):
        return bool(self._a) or bool(self._b)


def check_min_version(module, min_version):
    if isinstance(module, DeferredImportModule):
        indicator = module._indicator_flag
        indicator.resolve()
        if indicator._available:
            module = indicator._module
        else:
            return False

    version = getattr(module, '__version__', '0.0.0')
    return packaging.version.parse(min_version) <= packaging.version.parse(version)


#
# Note that we are duck-typing the Loader and MetaPathFinder base
# classes from importlib.abc.  This avoids a (surprisingly costly)
# import of importlib.abc
#
class DeferredImportCallbackLoader:
    """Custom Loader to resolve registered :py:class:`DeferredImportIndicator` objects

    This :py:class:`importlib.abc.Loader` loader wraps a regular loader
    and automatically resolves the registered
    :py:class:`DeferredImportIndicator` objects after the module is
    loaded.

    """

    def __init__(self, loader, deferred_indicators: List[DeferredImportIndicator]):
        self._loader = loader
        self._deferred_indicators = deferred_indicators

    def module_repr(self, module: ModuleType) -> str:
        return self._loader.module_repr(module)

    def create_module(self, spec) -> ModuleType:
        return self._loader.create_module(spec)

    def exec_module(self, module: ModuleType) -> None:
        self._loader.exec_module(module)
        # Now that the module has been loaded, trigger the resolution of
        # the deferred indicators (and their associated callbacks)
        for deferred in self._deferred_indicators:
            deferred.resolve()

    def load_module(self, fullname) -> ModuleType:
        return self._loader.load_module(fullname)

    def get_resource_reader(self, fullname):
        return self._loader.get_resource_reader(fullname)


class DeferredImportCallbackFinder:
    """Custom Finder that will wrap the normal loader to trigger callbacks

    This :py:class:`importlib.abc.MetaPathFinder` finder will wrap the
    normal loader returned by ``PathFinder`` with a loader that will
    trigger custom callbacks after the module is loaded.  We use this to
    trigger the post import callbacks registered through
    :py:func:`attempt_import` even when a user imports the target library
    directly (and not through attribute access on the
    :py:class:`DeferredImportModule`.

    """

    _callbacks = {}

    def find_spec(self, fullname, path, target=None):
        if fullname not in self._callbacks:
            return None

        spec = None
        # Continue looking for the finder that would have originally
        # loaded the deferred import module by starting at the next
        # finder in sys.meta_path (this way, we are agnostic to where
        # the module is coming from: file system, registry, etc.)
        for finder in sys.meta_path[sys.meta_path.index(self) + 1 :]:
            if hasattr(finder, 'find_spec'):
                # Support standard importlib MetaPathFinders
                spec = finder.find_spec(fullname, path, target)
                if spec is not None:
                    break
            else:
                # Support for imp finders/loaders (deprecated, but
                # supported through Python 3.11)
                loader = finder.find_module(fullname, path)
                if loader is not None:
                    spec = importlib.util.spec_from_loader(fullname, loader)
                    break
        else:
            # Module not found.  Returning None will proceed to the next
            # finder (which will eventually raise a ModuleNotFoundError)
            return None
        # Override the loader to trigger the finalization callback
        # after the original loader is finished
        spec.loader = DeferredImportCallbackLoader(
            spec.loader, self._callbacks[fullname]
        )
        return spec

    def invalidate_caches(self):
        pass


_DeferredImportCallbackFinder = DeferredImportCallbackFinder()
# Insert the DeferredImportCallbackFinder at the beginning of the
# sys.meta_path so that it is found before the standard finders (so that
# we can correctly inject the resolution of the DeferredImportIndicators
# -- which triggers the needed callbacks)
sys.meta_path.insert(0, _DeferredImportCallbackFinder)


def attempt_import(
    name,
    error_message=None,
    only_catch_importerror=None,
    minimum_version=None,
    alt_names=None,
    callback=None,
    importer=None,
    defer_check=None,
    defer_import=None,
    deferred_submodules=None,
    catch_exceptions=None,
):
    """Attempt to import the specified module.

    This will attempt to import the specified module, returning a
    ``(module, available)`` tuple.  If the import was successful, ``module``
    will be the imported module and ``available`` will be True.  If the
    import results in an exception, then ``module`` will be an instance of
    :py:class:`ModuleUnavailable` and ``available`` will be False

    The following

    .. doctest::

       >>> from pyomo.common.dependencies import attempt_import
       >>> numpy, numpy_available = attempt_import('numpy')

    Is roughly equivalent to

    .. doctest::

       >>> from pyomo.common.dependencies import ModuleUnavailable
       >>> try:
       ...     import numpy
       ...     numpy_available = True
       ... except ImportError as e:
       ...     numpy = ModuleUnavailable('numpy', 'Numpy is not available',
       ...                               '', str(e), globals()['__name__'])
       ...     numpy_available = False

    The import can be "deferred" until the first time the code either
    attempts to access the module or checks the Boolean value of the
    available flag.  This allows optional dependencies to be declared at
    the module scope but not imported until they are actually used by
    the module (thereby speeding up the initial package import).
    Deferred imports are handled by two helper classes
    (:py:class:`DeferredImportModule` and
    :py:class:`DeferredImportIndicator`).  Upon actual import,
    :py:meth:`DeferredImportIndicator.resolve()` attempts to replace
    those objects (in both the local and original global namespaces)
    with the imported module and Boolean flag so that subsequent uses of
    the module do not incur any overhead due to the delayed import.

    Parameters
    ----------
    name: str
        The name of the module to import

    error_message: str, optional
        The message for the exception raised by :py:class:`ModuleUnavailable`

    only_catch_importerror: bool, optional
        DEPRECATED: use ``catch_exceptions`` instead of ``only_catch_importerror``.

        If True (the default), exceptions other than ``ImportError`` raised
        during module import will be reraised.  If False, any exception
        will result in returning a :py:class:`ModuleUnavailable` object.
        (deprecated in version 5.7.3)

    minimum_version: str, optional
        The minimum acceptable module version (retrieved from
        ``module.__version__``)

    alt_names: list, optional
        DEPRECATED: ``alt_names`` no longer needs to be specified and is ignored.

        A list of common alternate names by which to look for this
        module in the ``globals()`` namespaces.  For example, the alt_names
        for NumPy would be ``['np']``.  (deprecated in version 6.0)

    callback: Callable[[ModuleType, bool], None], optional
        A function with the signature ``fcn(module, available)`` that
        will be called after the import is first attempted.

    importer: function, optional
        A function that will perform the import and return the imported
        module (or raise an :py:class:`ImportError`).  This is useful
        for cases where there are several equivalent modules and you
        want to import/return the first one that is available.

    defer_check: bool, optional
        DEPRECATED: renamed to ``defer_import`` (deprecated in version 6.7.2)

    defer_import: bool, optional
        If True, then the attempted import is deferred until the first
        use of either the module or the availability flag.  The method
        will return instances of :py:class:`DeferredImportModule` and
        :py:class:`DeferredImportIndicator`.  If False, the import will
        be attempted immediately.  If not set, then the import will be
        deferred unless the ``name`` is already present in
        ``sys.modules``.

    deferred_submodules: Iterable[str], optional
        If provided, an iterable of submodule names within this module
        that can be accessed without triggering a deferred import of
        this module.  For example, this module uses
        ``deferred_submodules=['pyplot', 'pylab']`` for ``matplotlib``.

    catch_exceptions: Iterable[Exception], optional
        If provided, this is the list of exceptions that will be caught
        when importing the target module, resulting in
        ``attempt_import`` returning a :py:class:`ModuleUnavailable`
        instance.  The default is to only catch :py:class:`ImportError`.
        This is useful when a module can regularly return additional
        exceptions during import.

    Returns
    -------
    : module
        the imported module, or an instance of
        :py:class:`ModuleUnavailable`, or an instance of
        :py:class:`DeferredImportModule`
    : bool
        Boolean indicating if the module import succeeded or an instance
        of :py:class:`DeferredImportIndicator`

    """
    if alt_names is not None:
        deprecation_warning(
            'alt_names=%s no longer needs to be specified '
            'and is ignored' % (alt_names,),
            version='6.0',
        )

    if only_catch_importerror is not None:
        deprecation_warning(
            "only_catch_importerror is deprecated.  Pass exceptions to "
            "catch using the catch_exceptions argument",
            version='5.7.3',
        )
        if catch_exceptions is not None:
            raise ValueError(
                "Cannot specify both only_catch_importerror and catch_exceptions"
            )
        if only_catch_importerror:
            catch_exceptions = (ImportError,)
        else:
            catch_exceptions = (ImportError, Exception)
    if catch_exceptions is None:
        catch_exceptions = (ImportError,)

    if defer_check is not None:
        deprecation_warning(
            'defer_check=%s is deprecated.  Please use defer_import' % (defer_check,),
            version='6.7.2',
        )
        assert defer_import is None
        defer_import = defer_check

    # If the module has already been imported, there is no reason to
    # further defer things: just import it.
    if defer_import is None:
        if name in sys.modules:
            defer_import = False
            deferred_submodules = None
        else:
            defer_import = True

    # If we are going to defer the check until later, return the
    # deferred import module object
    if defer_import:
        if deferred_submodules:
            if isinstance(deferred_submodules, Mapping):
                deprecation_warning(
                    'attempt_import(): deferred_submodules takes an iterable '
                    'and not a mapping (the alt_names supplied by the mapping '
                    'are no longer needed and are ignored).',
                    version='6.0',
                )
                deferred_submodules = list(deferred_submodules)

            # Ensures all names begin with '.'
            #
            # Fill in any missing submodules.  For example, if a user
            # provides ['foo.bar.baz'], then expand the list to
            # ['.foo', '.foo.bar', '.foo.bar.baz']
            deferred = []
            for _submod in deferred_submodules:
                if _submod[0] != '.':
                    _submod = '.' + _submod
                _mod_path = _submod.split('.')
                for i in range(len(_mod_path)):
                    _test_mod = '.'.join(_mod_path[:i])
                    if _test_mod not in deferred:
                        deferred.append(_test_mod)
                deferred.append(_submod)
            deferred = [_ for _ in deferred if _]
        else:
            deferred = None

        indicator = DeferredImportIndicator(
            name=name,
            error_message=error_message,
            catch_exceptions=catch_exceptions,
            minimum_version=minimum_version,
            original_globals=inspect.currentframe().f_back.f_globals,
            callback=callback,
            importer=importer,
            deferred_submodules=deferred,
        )
        return DeferredImportModule(indicator, deferred, None), indicator

    if deferred_submodules:
        raise ValueError("deferred_submodules is only valid if defer_import==True")

    return _perform_import(
        name=name,
        error_message=error_message,
        minimum_version=minimum_version,
        callback=callback,
        importer=importer,
        catch_exceptions=catch_exceptions,
        package=inspect.currentframe().f_back.f_globals.get('__name__', ''),
    )


def _perform_import(
    name, error_message, minimum_version, callback, importer, catch_exceptions, package
):
    import_error = None
    version_error = None
    try:
        with warnings.catch_warnings():
            # Temporarily suppress all warnings: we assume we are
            # importing a third-party package here and we don't want to
            # see them?
            if SUPPRESS_DEPENDENCY_WARNINGS and not name.startswith('pyomo.'):
                warnings.resetwarnings()
                warnings.simplefilter("ignore")
            if importer is None:
                module = importlib.import_module(name)
            else:
                module = importer()
        if minimum_version is None or check_min_version(module, minimum_version):
            if callback is not None:
                callback(module, True)
            return module, True
        else:
            version = getattr(module, '__version__', 'UNKNOWN')
            version_error = "version %s does not satisfy the minimum version %s" % (
                version,
                minimum_version,
            )
    except catch_exceptions as e:
        import_error = "%s: %s" % (type(e).__name__, e)

    module = ModuleUnavailable(
        name, error_message, version_error, import_error, package
    )
    if callback is not None:
        callback(module, False)
    return module, False


@deprecated(
    "``declare_deferred_modules_as_importable()`` is deprecated.  "
    "Use the :py:class:`declare_modules_as_importable` context manager.",
    version='6.7.2',
)
def declare_deferred_modules_as_importable(globals_dict):
    """Make all :py:class:`DeferredImportModules` in ``globals_dict`` importable

    This function will go throughout the specified ``globals_dict``
    dictionary and add any instances of :py:class:`DeferredImportModule`
    that it finds (and any of their deferred submodules) to
    ``sys.modules`` so that the modules can be imported through the
    ``globals_dict`` namespace.

    For example, ``pyomo/common/dependencies.py`` declares:

    .. doctest::
       :hide:

       >>> from pyomo.common.dependencies import (
       ...     attempt_import, _finalize_scipy, __dict__ as dep_globals,
       ...     declare_deferred_modules_as_importable, )
       >>> # Sphinx does not provide a proper globals()
       >>> def globals(): return dep_globals

    .. doctest::

       >>> scipy, scipy_available = attempt_import(
       ...     'scipy', callback=_finalize_scipy,
       ...     deferred_submodules=['stats', 'sparse', 'spatial', 'integrate'])
       >>> declare_deferred_modules_as_importable(globals())

    Which enables users to use:

    .. doctest::

       >>> import pyomo.common.dependencies.scipy.sparse as spa

    If the deferred import has not yet been triggered, then the
    :py:class:`DeferredImportModule` is returned and named ``spa``.
    However, if the import has already been triggered, then ``spa`` will
    either be the ``scipy.sparse`` module, or a
    :py:class:`ModuleUnavailable` instance.

    """
    return declare_modules_as_importable(globals_dict).__exit__(None, None, None)


class declare_modules_as_importable:
    """Make all :py:class:`ModuleType` and :py:class:`DeferredImportModules`
    importable through the ``globals_dict`` context.

    This context manager will detect all modules imported into the
    specified ``globals_dict`` environment (either directly or through
    :py:func:`attempt_import`) and will make those modules importable
    from the specified ``globals_dict`` context.  It works by detecting
    changes in the specified ``globals_dict`` dictionary and adding any new
    modules or instances of :py:class:`DeferredImportModule` that it
    finds (and any of their deferred submodules) to ``sys.modules`` so
    that the modules can be imported through the ``globals_dict``
    namespace.

    For example, ``pyomo/common/dependencies.py`` declares:

    .. doctest::
       :hide:

       >>> from pyomo.common.dependencies import (
       ...     attempt_import, _finalize_scipy, __dict__ as dep_globals,
       ...     declare_modules_as_importable, )
       >>> # Sphinx does not provide a proper globals()
       >>> def globals(): return dep_globals

    .. doctest::

       >>> with declare_modules_as_importable(globals()):
       ...     scipy, scipy_available = attempt_import(
       ...        'scipy', callback=_finalize_scipy,
       ...        deferred_submodules=['stats', 'sparse', 'spatial', 'integrate'])

    Which enables users to use:

    .. doctest::

       >>> import pyomo.common.dependencies.scipy.sparse as spa

    If the deferred import has not yet been triggered, then the
    :py:class:`DeferredImportModule` is returned and named ``spa``.
    However, if the import has already been triggered, then ``spa`` will
    either be the ``scipy.sparse`` module, or a
    :py:class:`ModuleUnavailable` instance.

    """

    def __init__(self, globals_dict):
        self.globals_dict = globals_dict
        self.init_dict = {}
        self.init_modules = None

    def __enter__(self):
        self.init_dict.update(self.globals_dict)
        self.init_modules = set(sys.modules)

    def __exit__(self, exc_type, exc_value, traceback):
        _global_name = self.globals_dict['__name__'] + '.'
        deferred = {
            k: v
            for k, v in self.globals_dict.items()
            if k not in self.init_dict
            and isinstance(v, (ModuleType, DeferredImportModule))
        }
        if self.init_modules:
            for name in set(sys.modules) - self.init_modules:
                if '.' in name and name.split('.', 1)[0] in deferred:
                    sys.modules[_global_name + name] = sys.modules[name]
        while deferred:
            name, mod = deferred.popitem()
            sys.modules[_global_name + name] = mod
            if isinstance(mod, DeferredImportModule):
                mod.__path__ = None
                mod.__spec__ = None
                deferred.update(
                    (name + '.' + k, v)
                    for k, v in mod.__dict__.items()
                    if type(v) is DeferredImportModule
                )


#
# Common optional dependencies used throughout Pyomo
#

yaml_load_args = {}


def _finalize_yaml(module, available):
    # Recent versions of PyYAML issue warnings if the Loader argument is
    # not set
    if available and hasattr(module, 'SafeLoader'):
        yaml_load_args['Loader'] = module.SafeLoader


def _finalize_ctypes(module, available):
    # ctypes.util must be explicitly imported (and fileutils assumes
    # this has already happened)
    import ctypes.util


def _finalize_scipy(module, available):
    if available:
        # Import key subpackages that we will want to assume are present
        import scipy.stats

        # As of scipy 1.6.0, importing scipy.stats causes the following
        # to be automatically imported.  However, we will still
        # explicitly import them here to guard against potential future
        # changes in scipy.
        import scipy.integrate
        import scipy.sparse
        import scipy.spatial


def _finalize_pympler(module, available):
    if available:
        # Import key subpackages that we will want to assume are present
        import pympler.muppy


def _finalize_packaging(module, available):
    if available:
        # Import key subpackages that we will want to assume are present
        import packaging.version


def _finalize_matplotlib(module, available):
    if not available:
        return
    # You must switch matplotlib backends *before* importing pyplot.  If
    # we are in the middle of testing, we need to switch the backend to
    # 'Agg', otherwise attempts to generate plots on CI services without
    # terminal windows will fail.
    if in_testing_environment():
        module.use('Agg')
    import matplotlib.pyplot
    import matplotlib.pylab
    import matplotlib.backends


def _finalize_mpi4py(module, available):
    if not available:
        return
    import mpi4py.MPI


def _finalize_numpy(np, available):
    if not available:
        return
    # scipy has a dependence on numpy.testing, and if we don't import it
    # as part of resolving numpy, then certain deferred scipy imports
    # fail when run under pytest.
    import numpy.testing

    from . import numeric_types

    # Register ndarray as a native type to prevent 1-element ndarrays
    # from accidentally registering ndarray as a native_numeric_type.
    numeric_types.native_types.add(np.ndarray)
    numeric_types.RegisterLogicalType(np.bool_)
    for t in (
        np.int_,
        np.intc,
        np.intp,
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.uint8,
        np.uint16,
        np.uint32,
        np.uint64,
    ):
        numeric_types.RegisterIntegerType(t)
        # We have deprecated RegisterBooleanType, so we will mock up the
        # registration here (to bypass the deprecation warning) until we
        # finally remove all support for it
        numeric_types._native_boolean_types.add(t)
    _floats = [np.float16, np.float32, np.float64]
    # float96 and float128 may or may not be defined in this particular
    # numpy build (it depends on platform and version).
    # Register them only if they are present
    if hasattr(np, 'float_'):
        # Prepend to preserve previous functionality
        _floats.insert(0, np.float_)
    if hasattr(np, 'float96'):
        _floats.append(np.float96)
    if hasattr(np, 'float128'):
        _floats.append(np.float128)
    for t in _floats:
        numeric_types.RegisterNumericType(t)
        # We have deprecated RegisterBooleanType, so we will mock up the
        # registration here (to bypass the deprecation warning) until we
        # finally remove all support for it
        numeric_types._native_boolean_types.add(t)
    _complex = [np.complex64, np.complex128]
    # complex192 and complex256 may or may not be defined in this
    # particular numpy build (it depends on platform and version).
    # Register them only if they are present
    if hasattr(np, 'np.complex_'):
        # Prepend to preserve functionality
        _complex.insert(0, np.complex_)
    if hasattr(np, 'complex192'):
        _complex.append(np.complex192)
    if hasattr(np, 'complex256'):
        _complex.append(np.complex256)
    for t in _complex:
        numeric_types.RegisterComplexType(t)


def _pyutilib_importer():
    # On newer Pythons, PyUtilib import will fail, but only if a
    # second-level module is imported.  We will arbitrarily choose to
    # check pyutilib.component (as that is the path exercised by the
    # pyomo.common.tempfiles deprecation path)
    importlib.import_module('pyutilib.component')
    return importlib.import_module('pyutilib')


with declare_modules_as_importable(globals()):
    # Standard libraries that are slower to import and not strictly required
    # on all platforms / situations.
    ctypes, _ = attempt_import(
        'ctypes', deferred_submodules=['util'], callback=_finalize_ctypes
    )
    multiprocessing, _ = attempt_import('multiprocessing')
    random, _ = attempt_import('random')

    # Necessary for minimum version checking for other optional dependencies
    packaging, packaging_available = attempt_import(
        'packaging', deferred_submodules=['version'], callback=_finalize_packaging
    )
    # Commonly-used optional dependencies
    dill, dill_available = attempt_import('dill')
    mpi4py, mpi4py_available = attempt_import(
        'mpi4py', deferred_submodules=['MPI'], callback=_finalize_mpi4py
    )
    networkx, networkx_available = attempt_import('networkx')
    numpy, numpy_available = attempt_import('numpy', callback=_finalize_numpy)
    pandas, pandas_available = attempt_import('pandas')
    pathlib, pathlib_available = attempt_import('pathlib')
    pint, pint_available = attempt_import(
        'pint',
        # TypeError for pint<=0.24.3 and python>=3.13
        catch_exceptions=(ImportError, TypeError),
    )
    plotly, plotly_available = attempt_import('plotly')
    pympler, pympler_available = attempt_import('pympler', callback=_finalize_pympler)
    pyutilib, pyutilib_available = attempt_import(
        'pyutilib', importer=_pyutilib_importer
    )
    scipy, scipy_available = attempt_import(
        'scipy',
        callback=_finalize_scipy,
        deferred_submodules=['stats', 'sparse', 'spatial', 'integrate'],
    )
    yaml, yaml_available = attempt_import('yaml', callback=_finalize_yaml)

    # Note that matplotlib.pyplot can generate a runtime error on OSX when
    # not installed as a Framework (as is the case in the CI systems)
    matplotlib, matplotlib_available = attempt_import(
        'matplotlib',
        callback=_finalize_matplotlib,
        deferred_submodules=['pyplot', 'pylab', 'backends'],
        catch_exceptions=(ImportError, RuntimeError),
    )

try:
    import cPickle as pickle
except ImportError:
    import pickle



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.pynumero.interfaces.utils import (
    build_bounds_mask,
    build_compression_matrix,
)
import numpy as np
import logging
import time
from pyomo.contrib.pynumero.linalg.base import LinearSolverStatus
from pyomo.common.timing import HierarchicalTimer
import enum

"""
Interface Requirements
----------------------
1) duals_primals_lb[i] must always be 0 if primals_lb[i] is -inf
2) duals_primals_ub[i] must always be 0 if primals_ub[i] is inf
3) duals_slacks_lb[i] must always be 0 if ineq_lb[i] is -inf
4) duals_slacks_ub[i] must always be 0 if ineq_ub[i] is inf
"""


ip_logger = logging.getLogger('interior_point')


class InteriorPointStatus(enum.Enum):
    optimal = 0
    error = 1


class LinearSolveContext:
    def __init__(
        self,
        interior_point_logger,
        linear_solver_logger,
        filename=None,
        level=logging.INFO,
    ):
        self.interior_point_logger = interior_point_logger
        self.linear_solver_logger = linear_solver_logger
        self.filename = filename

        if filename:
            self.handler = logging.FileHandler(filename)
            self.handler.setLevel(level)

    def __enter__(self):
        self.linear_solver_logger.propagate = False
        self.interior_point_logger.propagate = False
        if self.filename:
            self.linear_solver_logger.addHandler(self.handler)
            self.interior_point_logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        self.linear_solver_logger.propagate = True
        self.interior_point_logger.propagate = True
        if self.filename:
            self.linear_solver_logger.removeHandler(self.handler)
            self.interior_point_logger.removeHandler(self.handler)


# How should the RegContext work?
# TODO: in this class, use the linear_solver_context to ...
#       Use linear_solver_logger to write iter_no and reg_coef
#
#       Define a method for logging IP_reg_info to the linear solver log
#       Method can be called within linear_solve_context
class FactorizationContext:
    def __init__(self, logger):
        # Any reason to pass in a logging level here?
        # ^ So the "regularization log" can have its own outlvl
        self.logger = logger

    def __enter__(self):
        self.logger.debug('Factorizing KKT')
        self.log_header()
        return self

    def __exit__(self, et, ev, tb):
        self.logger.debug('Finished factorizing KKT')
        # Will this swallow exceptions in this context?

    def log_header(self):
        self.logger.debug(
            '{_iter:<10}'
            '{reg_iter:<10}'
            '{num_realloc:<10}'
            '{reg_coef:<10}'
            '{neg_eig:<10}'
            '{status:<10}'.format(
                _iter='Iter',
                reg_iter='reg_iter',
                num_realloc='# realloc',
                reg_coef='reg_coef',
                neg_eig='neg_eig',
                status='status',
            )
        )

    def log_info(self, _iter, reg_iter, num_realloc, coef, neg_eig, status):
        self.logger.debug(
            '{_iter:<10}'
            '{reg_iter:<10}'
            '{num_realloc:<10}'
            '{reg_coef:<10.2e}'
            '{neg_eig:<10}'
            '{status:<10}'.format(
                _iter=_iter,
                reg_iter=reg_iter,
                num_realloc=num_realloc,
                reg_coef=coef,
                neg_eig=str(neg_eig),
                status=status.name,
            )
        )


class InteriorPointSolver:
    """
    Class for creating interior point solvers with different options
    """

    def __init__(
        self,
        linear_solver,
        max_iter=100,
        tol=1e-8,
        linear_solver_log_filename=None,
        max_reallocation_iterations=5,
        reallocation_factor=2,
    ):
        self.linear_solver = linear_solver
        self.max_iter = max_iter
        self.tol = tol
        self.linear_solver_log_filename = linear_solver_log_filename
        self.max_reallocation_iterations = max_reallocation_iterations
        self.reallocation_factor = reallocation_factor
        self.base_eq_reg_coef = -1e-8
        self._barrier_parameter = 0.1
        self._minimum_barrier_parameter = 1e-9
        self.hess_reg_coef = 1e-4
        self.max_reg_iter = 6
        self.reg_factor_increase = 100

        self.logger = logging.getLogger('interior_point')
        self._iter = 0
        self.factorization_context = FactorizationContext(self.logger)

        if linear_solver_log_filename:
            with open(linear_solver_log_filename, 'w'):
                pass

        self.linear_solver_logger = self.linear_solver.getLogger()
        self.linear_solve_context = LinearSolveContext(
            self.logger, self.linear_solver_logger, self.linear_solver_log_filename
        )

    def update_barrier_parameter(self):
        self._barrier_parameter = max(
            self._minimum_barrier_parameter,
            min(0.5 * self._barrier_parameter, self._barrier_parameter**1.5),
        )

    def set_linear_solver(self, linear_solver):
        """This method exists to hopefully make it easy to try the same IP
        algorithm with different linear solvers.
        Subclasses may have linear-solver specific methods, in which case
        this should not be called.

        Hopefully the linear solver interface can be standardized such that
        this is not a problem. (Need a generalized method for set_options)
        """
        self.linear_solver = linear_solver

    def set_interface(self, interface):
        self.interface = interface

    def solve(self, interface, timer=None, report_timing=False):
        """
        Parameters
        ----------
        interface: pyomo.contrib.interior_point.interface.BaseInteriorPointInterface
            The interior point interface. This object handles the function evaluation,
            building the KKT matrix, and building the KKT right hand side.
        timer: HierarchicalTimer
        report_timing: bool
        """
        linear_solver = self.linear_solver
        max_iter = self.max_iter
        tol = self.tol
        if timer is None:
            timer = HierarchicalTimer()

        timer.start('IP solve')
        timer.start('init')

        self._barrier_parameter = 0.1

        self.set_interface(interface)

        t0 = time.time()
        primals = interface.init_primals().copy()
        slacks = interface.init_slacks().copy()
        duals_eq = interface.init_duals_eq().copy()
        duals_ineq = interface.init_duals_ineq().copy()
        duals_primals_lb = interface.init_duals_primals_lb().copy()
        duals_primals_ub = interface.init_duals_primals_ub().copy()
        duals_slacks_lb = interface.init_duals_slacks_lb().copy()
        duals_slacks_ub = interface.init_duals_slacks_ub().copy()

        self.process_init(primals, interface.primals_lb(), interface.primals_ub())
        self.process_init(slacks, interface.ineq_lb(), interface.ineq_ub())
        self.process_init_duals_lb(duals_primals_lb, self.interface.primals_lb())
        self.process_init_duals_ub(duals_primals_ub, self.interface.primals_ub())
        self.process_init_duals_lb(duals_slacks_lb, self.interface.ineq_lb())
        self.process_init_duals_ub(duals_slacks_ub, self.interface.ineq_ub())

        interface.set_barrier_parameter(self._barrier_parameter)

        alpha_primal_max = 1
        alpha_dual_max = 1

        self.logger.info(
            '{_iter:<6}'
            '{objective:<11}'
            '{primal_inf:<11}'
            '{dual_inf:<11}'
            '{compl_inf:<11}'
            '{barrier:<11}'
            '{alpha_p:<11}'
            '{alpha_d:<11}'
            '{reg:<11}'
            '{time:<7}'.format(
                _iter='Iter',
                objective='Objective',
                primal_inf='Prim Inf',
                dual_inf='Dual Inf',
                compl_inf='Comp Inf',
                barrier='Barrier',
                alpha_p='Prim Step',
                alpha_d='Dual Step',
                reg='Reg',
                time='Time',
            )
        )

        reg_coef = 0

        timer.stop('init')
        status = InteriorPointStatus.error

        for _iter in range(max_iter):
            self._iter = _iter

            interface.set_primals(primals)
            interface.set_slacks(slacks)
            interface.set_duals_eq(duals_eq)
            interface.set_duals_ineq(duals_ineq)
            interface.set_duals_primals_lb(duals_primals_lb)
            interface.set_duals_primals_ub(duals_primals_ub)
            interface.set_duals_slacks_lb(duals_slacks_lb)
            interface.set_duals_slacks_ub(duals_slacks_ub)

            timer.start('convergence check')
            primal_inf, dual_inf, complimentarity_inf = self.check_convergence(
                barrier=0, timer=timer
            )
            timer.stop('convergence check')
            objective = interface.evaluate_objective()
            self.logger.info(
                '{_iter:<6}'
                '{objective:<11.2e}'
                '{primal_inf:<11.2e}'
                '{dual_inf:<11.2e}'
                '{compl_inf:<11.2e}'
                '{barrier:<11.2e}'
                '{alpha_p:<11.2e}'
                '{alpha_d:<11.2e}'
                '{reg:<11.2e}'
                '{time:<7.3f}'.format(
                    _iter=_iter,
                    objective=objective,
                    primal_inf=primal_inf,
                    dual_inf=dual_inf,
                    compl_inf=complimentarity_inf,
                    barrier=self._barrier_parameter,
                    alpha_p=alpha_primal_max,
                    alpha_d=alpha_dual_max,
                    reg=reg_coef,
                    time=time.time() - t0,
                )
            )

            if max(primal_inf, dual_inf, complimentarity_inf) <= tol:
                status = InteriorPointStatus.optimal
                break
            timer.start('convergence check')
            primal_inf, dual_inf, complimentarity_inf = self.check_convergence(
                barrier=self._barrier_parameter, timer=timer
            )
            timer.stop('convergence check')
            if (
                max(primal_inf, dual_inf, complimentarity_inf)
                <= 0.1 * self._barrier_parameter
            ):
                # This comparison is made with barrier problem infeasibility.
                # Sometimes have trouble getting dual infeasibility low enough
                self.update_barrier_parameter()

            interface.set_barrier_parameter(self._barrier_parameter)
            timer.start('eval')
            timer.start('eval kkt')
            kkt = interface.evaluate_primal_dual_kkt_matrix(timer=timer)
            timer.stop('eval kkt')
            timer.start('eval rhs')
            rhs = interface.evaluate_primal_dual_kkt_rhs(timer=timer)
            timer.stop('eval rhs')
            timer.stop('eval')

            # Factorize linear system
            timer.start('factorize')
            reg_coef = self.factorize(kkt=kkt, timer=timer)
            timer.stop('factorize')

            timer.start('back solve')
            with self.linear_solve_context:
                self.logger.info('Iter: %s' % self._iter)
                delta, res = linear_solver.do_back_solve(rhs)
                if res.status != LinearSolverStatus.successful:
                    raise RuntimeError(f'Backsolve failed: {res.status}')
            timer.stop('back solve')

            interface.set_primal_dual_kkt_solution(delta)
            timer.start('frac boundary')
            alpha_primal_max, alpha_dual_max = self.fraction_to_the_boundary()
            timer.stop('frac boundary')
            delta_primals = interface.get_delta_primals()
            delta_slacks = interface.get_delta_slacks()
            delta_duals_eq = interface.get_delta_duals_eq()
            delta_duals_ineq = interface.get_delta_duals_ineq()
            delta_duals_primals_lb = interface.get_delta_duals_primals_lb()
            delta_duals_primals_ub = interface.get_delta_duals_primals_ub()
            delta_duals_slacks_lb = interface.get_delta_duals_slacks_lb()
            delta_duals_slacks_ub = interface.get_delta_duals_slacks_ub()

            primals += alpha_primal_max * delta_primals
            slacks += alpha_primal_max * delta_slacks
            duals_eq += alpha_dual_max * delta_duals_eq
            duals_ineq += alpha_dual_max * delta_duals_ineq
            duals_primals_lb += alpha_dual_max * delta_duals_primals_lb
            duals_primals_ub += alpha_dual_max * delta_duals_primals_ub
            duals_slacks_lb += alpha_dual_max * delta_duals_slacks_lb
            duals_slacks_ub += alpha_dual_max * delta_duals_slacks_ub

        timer.stop('IP solve')
        if report_timing:
            print(timer)
        return status

    def factorize(self, kkt, timer=None):
        desired_n_neg_evals = (
            self.interface.n_eq_constraints() + self.interface.n_ineq_constraints()
        )
        reg_iter = 0
        with self.factorization_context as fact_con:
            status, num_realloc = try_factorization_and_reallocation(
                kkt=kkt,
                linear_solver=self.linear_solver,
                reallocation_factor=self.reallocation_factor,
                max_iter=self.max_reallocation_iterations,
                timer=timer,
            )
            if status not in {
                LinearSolverStatus.successful,
                LinearSolverStatus.singular,
            }:
                raise RuntimeError(
                    'Could not factorize KKT system; linear solver status: '
                    + str(status)
                )

            if status == LinearSolverStatus.successful:
                neg_eig = self.linear_solver.get_inertia()[1]
            else:
                neg_eig = None
            fact_con.log_info(
                _iter=self._iter,
                reg_iter=reg_iter,
                num_realloc=num_realloc,
                coef=0,
                neg_eig=neg_eig,
                status=status,
            )
            reg_iter += 1

            if status == LinearSolverStatus.singular:
                constraint_reg_coef = (
                    self.base_eq_reg_coef * self._barrier_parameter**0.25
                )
                kkt = self.interface.regularize_equality_gradient(
                    kkt=kkt, coef=constraint_reg_coef, copy_kkt=False
                )

            total_hess_reg_coef = self.hess_reg_coef
            last_hess_reg_coef = 0

            while (
                neg_eig != desired_n_neg_evals or status == LinearSolverStatus.singular
            ):
                kkt = self.interface.regularize_hessian(
                    kkt=kkt,
                    coef=total_hess_reg_coef - last_hess_reg_coef,
                    copy_kkt=False,
                )
                status, num_realloc = try_factorization_and_reallocation(
                    kkt=kkt,
                    linear_solver=self.linear_solver,
                    reallocation_factor=self.reallocation_factor,
                    max_iter=self.max_reallocation_iterations,
                    timer=timer,
                )
                if status != LinearSolverStatus.successful:
                    raise RuntimeError(
                        'Could not factorize KKT system; linear solver status: '
                        + str(status)
                    )
                neg_eig = self.linear_solver.get_inertia()[1]
                fact_con.log_info(
                    _iter=self._iter,
                    reg_iter=reg_iter,
                    num_realloc=num_realloc,
                    coef=total_hess_reg_coef,
                    neg_eig=neg_eig,
                    status=status,
                )
                reg_iter += 1
                if reg_iter > self.max_reg_iter:
                    raise RuntimeError(
                        'Exceeded maximum number of regularization iterations.'
                    )
                last_hess_reg_coef = total_hess_reg_coef
                total_hess_reg_coef *= self.reg_factor_increase

        return last_hess_reg_coef

    def process_init(self, x, lb, ub):
        process_init(x, lb, ub)

    def process_init_duals_lb(self, x, lb):
        process_init_duals_lb(x, lb)

    def process_init_duals_ub(self, x, ub):
        process_init_duals_ub(x, ub)

    def check_convergence(self, barrier, timer=None):
        """
        Parameters
        ----------
        barrier: float
        timer: HierarchicalTimer

        Returns
        -------
        primal_inf: float
        dual_inf: float
        complimentarity_inf: float
        """
        if timer is None:
            timer = HierarchicalTimer()

        interface = self.interface
        slacks = interface.get_slacks()
        timer.start('grad obj')
        grad_obj = interface.get_obj_factor() * interface.evaluate_grad_objective()
        timer.stop('grad obj')
        timer.start('jac eq')
        jac_eq = interface.evaluate_jacobian_eq()
        timer.stop('jac eq')
        timer.start('jac ineq')
        jac_ineq = interface.evaluate_jacobian_ineq()
        timer.stop('jac ineq')
        timer.start('eq cons')
        eq_resid = interface.evaluate_eq_constraints()
        timer.stop('eq cons')
        timer.start('ineq cons')
        ineq_resid = interface.evaluate_ineq_constraints() - slacks
        timer.stop('ineq cons')
        primals = interface.get_primals()
        duals_eq = interface.get_duals_eq()
        duals_ineq = interface.get_duals_ineq()
        duals_primals_lb = interface.get_duals_primals_lb()
        duals_primals_ub = interface.get_duals_primals_ub()
        duals_slacks_lb = interface.get_duals_slacks_lb()
        duals_slacks_ub = interface.get_duals_slacks_ub()

        primals_lb = interface.primals_lb()
        primals_ub = interface.primals_ub()
        primals_lb_mod = primals_lb.copy()
        primals_ub_mod = primals_ub.copy()
        primals_lb_mod[np.isneginf(primals_lb)] = 0  # these entries get multiplied by 0
        primals_ub_mod[np.isinf(primals_ub)] = 0  # these entries get multiplied by 0

        ineq_lb = interface.ineq_lb()
        ineq_ub = interface.ineq_ub()
        ineq_lb_mod = ineq_lb.copy()
        ineq_ub_mod = ineq_ub.copy()
        ineq_lb_mod[np.isneginf(ineq_lb)] = 0  # these entries get multiplied by 0
        ineq_ub_mod[np.isinf(ineq_ub)] = 0  # these entries get multiplied by 0

        timer.start('grad_lag_primals')
        grad_lag_primals = grad_obj + jac_eq.transpose() * duals_eq
        grad_lag_primals += jac_ineq.transpose() * duals_ineq
        grad_lag_primals -= duals_primals_lb
        grad_lag_primals += duals_primals_ub
        timer.stop('grad_lag_primals')
        timer.start('grad_lag_slacks')
        grad_lag_slacks = -duals_ineq - duals_slacks_lb + duals_slacks_ub
        timer.stop('grad_lag_slacks')
        timer.start('bound resids')
        primals_lb_resid = (primals - primals_lb_mod) * duals_primals_lb - barrier
        primals_ub_resid = (primals_ub_mod - primals) * duals_primals_ub - barrier
        primals_lb_resid[np.isneginf(primals_lb)] = 0
        primals_ub_resid[np.isinf(primals_ub)] = 0
        slacks_lb_resid = (slacks - ineq_lb_mod) * duals_slacks_lb - barrier
        slacks_ub_resid = (ineq_ub_mod - slacks) * duals_slacks_ub - barrier
        slacks_lb_resid[np.isneginf(ineq_lb)] = 0
        slacks_ub_resid[np.isinf(ineq_ub)] = 0
        timer.stop('bound resids')

        if eq_resid.size == 0:
            max_eq_resid = 0
        else:
            max_eq_resid = np.max(np.abs(eq_resid))
        if ineq_resid.size == 0:
            max_ineq_resid = 0
        else:
            max_ineq_resid = np.max(np.abs(ineq_resid))
        primal_inf = max(max_eq_resid, max_ineq_resid)

        max_grad_lag_primals = np.max(np.abs(grad_lag_primals))
        if grad_lag_slacks.size == 0:
            max_grad_lag_slacks = 0
        else:
            max_grad_lag_slacks = np.max(np.abs(grad_lag_slacks))
        dual_inf = max(max_grad_lag_primals, max_grad_lag_slacks)

        if primals_lb_resid.size == 0:
            max_primals_lb_resid = 0
        else:
            max_primals_lb_resid = np.max(np.abs(primals_lb_resid))
        if primals_ub_resid.size == 0:
            max_primals_ub_resid = 0
        else:
            max_primals_ub_resid = np.max(np.abs(primals_ub_resid))
        if slacks_lb_resid.size == 0:
            max_slacks_lb_resid = 0
        else:
            max_slacks_lb_resid = np.max(np.abs(slacks_lb_resid))
        if slacks_ub_resid.size == 0:
            max_slacks_ub_resid = 0
        else:
            max_slacks_ub_resid = np.max(np.abs(slacks_ub_resid))
        complimentarity_inf = max(
            max_primals_lb_resid,
            max_primals_ub_resid,
            max_slacks_lb_resid,
            max_slacks_ub_resid,
        )

        return primal_inf, dual_inf, complimentarity_inf

    def fraction_to_the_boundary(self):
        return fraction_to_the_boundary(self.interface, 1 - self._barrier_parameter)


def try_factorization_and_reallocation(
    kkt, linear_solver, reallocation_factor, max_iter, timer=None
):
    if timer is None:
        timer = HierarchicalTimer()

    assert max_iter >= 1
    for count in range(max_iter):
        timer.start('symbolic')
        """
        Performance could be improved significantly by only performing 
        symbolic factorization once.

        However, we first have to make sure the nonzero structure 
        (and ordering of row and column arrays) of the KKT matrix never 
        changes. We have not had time to test this thoroughly, yet. 
        """
        res = linear_solver.do_symbolic_factorization(matrix=kkt, raise_on_error=False)
        timer.stop('symbolic')
        if res.status == LinearSolverStatus.successful:
            timer.start('numeric')
            res = linear_solver.do_numeric_factorization(
                matrix=kkt, raise_on_error=False
            )
            timer.stop('numeric')
        status = res.status
        if status == LinearSolverStatus.not_enough_memory:
            linear_solver.increase_memory_allocation(reallocation_factor)
        else:
            break
    return status, count


def _fraction_to_the_boundary_helper_lb(tau, x, delta_x, xl):
    delta_x_mod = delta_x.copy()
    delta_x_mod[delta_x_mod == 0] = 1
    alpha = -tau * (x - xl) / delta_x_mod
    alpha[delta_x >= 0] = np.inf
    if alpha.size == 0:
        return 1
    else:
        return min(alpha.min(), 1)


def _fraction_to_the_boundary_helper_ub(tau, x, delta_x, xu):
    delta_x_mod = delta_x.copy()
    delta_x_mod[delta_x_mod == 0] = 1
    alpha = tau * (xu - x) / delta_x_mod
    alpha[delta_x <= 0] = np.inf
    if alpha.size == 0:
        return 1
    else:
        return min(alpha.min(), 1)


def fraction_to_the_boundary(interface, tau):
    """
    Parameters
    ----------
    interface: pyomo.contrib.interior_point.interface.BaseInteriorPointInterface
    tau: float

    Returns
    -------
    alpha_primal_max: float
    alpha_dual_max: float
    """
    primals = interface.get_primals()
    slacks = interface.get_slacks()
    duals_primals_lb = interface.get_duals_primals_lb()
    duals_primals_ub = interface.get_duals_primals_ub()
    duals_slacks_lb = interface.get_duals_slacks_lb()
    duals_slacks_ub = interface.get_duals_slacks_ub()

    delta_primals = interface.get_delta_primals()
    delta_slacks = interface.get_delta_slacks()
    delta_duals_primals_lb = interface.get_delta_duals_primals_lb()
    delta_duals_primals_ub = interface.get_delta_duals_primals_ub()
    delta_duals_slacks_lb = interface.get_delta_duals_slacks_lb()
    delta_duals_slacks_ub = interface.get_delta_duals_slacks_ub()

    primals_lb = interface.primals_lb()
    primals_ub = interface.primals_ub()
    ineq_lb = interface.ineq_lb()
    ineq_ub = interface.ineq_ub()

    alpha_primal_max_a = _fraction_to_the_boundary_helper_lb(
        tau=tau, x=primals, delta_x=delta_primals, xl=primals_lb
    )
    alpha_primal_max_b = _fraction_to_the_boundary_helper_ub(
        tau=tau, x=primals, delta_x=delta_primals, xu=primals_ub
    )
    alpha_primal_max_c = _fraction_to_the_boundary_helper_lb(
        tau=tau, x=slacks, delta_x=delta_slacks, xl=ineq_lb
    )
    alpha_primal_max_d = _fraction_to_the_boundary_helper_ub(
        tau=tau, x=slacks, delta_x=delta_slacks, xu=ineq_ub
    )
    alpha_primal_max = min(
        alpha_primal_max_a, alpha_primal_max_b, alpha_primal_max_c, alpha_primal_max_d
    )

    alpha_dual_max_a = _fraction_to_the_boundary_helper_lb(
        tau=tau,
        x=duals_primals_lb,
        delta_x=delta_duals_primals_lb,
        xl=np.zeros(duals_primals_lb.size),
    )
    alpha_dual_max_b = _fraction_to_the_boundary_helper_lb(
        tau=tau,
        x=duals_primals_ub,
        delta_x=delta_duals_primals_ub,
        xl=np.zeros(duals_primals_ub.size),
    )
    alpha_dual_max_c = _fraction_to_the_boundary_helper_lb(
        tau=tau,
        x=duals_slacks_lb,
        delta_x=delta_duals_slacks_lb,
        xl=np.zeros(duals_slacks_lb.size),
    )
    alpha_dual_max_d = _fraction_to_the_boundary_helper_lb(
        tau=tau,
        x=duals_slacks_ub,
        delta_x=delta_duals_slacks_ub,
        xl=np.zeros(duals_slacks_ub.size),
    )
    alpha_dual_max = min(
        alpha_dual_max_a, alpha_dual_max_b, alpha_dual_max_c, alpha_dual_max_d
    )

    return alpha_primal_max, alpha_dual_max


def process_init(x, lb, ub):
    if np.any((ub - lb) < 0):
        raise ValueError(
            'Lower bounds for variables/inequalities should not be larger '
            'than upper bounds.'
        )
    if np.any((ub - lb) == 0):
        raise ValueError(
            'Variables and inequalities should not have equal lower and upper '
            'bounds.'
        )

    lb_mask = build_bounds_mask(lb)
    ub_mask = build_bounds_mask(ub)

    lb_only = np.logical_and(lb_mask, np.logical_not(ub_mask))
    ub_only = np.logical_and(ub_mask, np.logical_not(lb_mask))
    lb_and_ub = np.logical_and(lb_mask, ub_mask)
    out_of_bounds = (x >= ub) + (x <= lb)
    out_of_bounds_lb_only = np.logical_and(out_of_bounds, lb_only)
    out_of_bounds_ub_only = np.logical_and(out_of_bounds, ub_only)
    out_of_bounds_lb_and_ub = np.logical_and(out_of_bounds, lb_and_ub)

    cm = build_compression_matrix(out_of_bounds_lb_only)
    x[out_of_bounds_lb_only] = cm * (lb + 1)

    cm = build_compression_matrix(out_of_bounds_ub_only)
    x[out_of_bounds_ub_only] = cm * (ub - 1)

    del cm
    cm1 = build_compression_matrix(lb_and_ub)
    cm2 = build_compression_matrix(out_of_bounds_lb_and_ub)
    x[out_of_bounds_lb_and_ub] = cm2 * (0.5 * cm1.transpose() * (cm1 * lb + cm1 * ub))


def process_init_duals_lb(x, lb):
    x[x <= 0] = 1
    x[np.isneginf(lb)] = 0


def process_init_duals_ub(x, ub):
    x[x <= 0] = 1
    x[np.isinf(ub)] = 0



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

from pyomo.core.base.componentuid import ComponentUID
from pyomo.core.base.expression import Expression
from pyomo.core.base.set import Set

from pyomo.contrib.mpc.data.series_data import get_indexed_cuid
from pyomo.contrib.mpc.data.scalar_data import ScalarData


def _get_quadratic_penalty_at_time(var, t, target, weight=None):
    if weight is None:
        weight = 1.0
    return weight * (var[t] - target) ** 2


def _get_penalty_expressions_at_time(
    variables, t, target_data, weight_data=None, time_set=None
):
    """A private helper function to process data and construct penalty
    expressions

    """
    if weight_data is None:
        weight_data = ScalarData(ComponentMap((var, 1.0) for var in variables))
    if not isinstance(weight_data, ScalarData):
        # We pass time_set as an argument in case the user provides a
        # ComponentMap of VarData -> values. In this case knowing the
        # time set is necessary to recover the indexed CUID.
        weight_data = ScalarData(weight_data, time_set=time_set)
    if not isinstance(target_data, ScalarData):
        target_data = ScalarData(target_data, time_set=time_set)

    for var in variables:
        if not target_data.contains_key(var):
            raise KeyError(
                "Target data does not contain a key for variable %s" % var.name
            )
        if not weight_data.contains_key(var):
            raise KeyError(
                "Penalty weight data does not contain a key for variable %s" % var.name
            )

    penalties = [
        _get_quadratic_penalty_at_time(
            var,
            t,
            target_data.get_data_from_key(var),
            weight_data.get_data_from_key(var),
        )
        for var in variables
    ]
    return penalties


def get_penalty_at_time(
    variables, t, target_data, weight_data=None, time_set=None, variable_set=None
):
    """Returns an Expression penalizing the deviation of the specified
    variables at the specified point in time from the specified target

    Arguments
    ---------
    variables: List
        List of time-indexed variables that will be penalized
    t: Float
        Time point at which to apply the penalty
    target_data: ~scalar_data.ScalarData
        ScalarData object containing the target for (at least) the variables
        to be penalized
    weight_data: ~scalar_data.ScalarData (optional)
        ScalarData object containing the penalty weights for (at least) the
        variables to be penalized
    time_set: Set (optional)
        Time set that indexes the provided variables. This is only used if
        target or weight data are provided as a ComponentMap with VarData
        as keys. In this case the Set is necessary to recover the CUIDs
        used internally as keys
    variable_set: Set (optional)
        Set indexing the list of variables provided, if such a set already
        exists

    Returns
    -------
    Set, Expression
        Set indexing the list of variables provided and an Expression,
        indexed by this set, containing the weighted penalty expressions

    """
    if variable_set is None:
        variable_set = Set(initialize=range(len(variables)))
    penalty_expressions = _get_penalty_expressions_at_time(
        variables, t, target_data, weight_data=weight_data, time_set=time_set
    )

    def penalty_rule(m, i):
        return penalty_expressions[i]

    penalty = Expression(variable_set, rule=penalty_rule)
    return variable_set, penalty


def get_terminal_penalty(
    variables, time_set, target_data, weight_data=None, variable_set=None
):
    """Returns an Expression penalizing the deviation of the specified
    variables at the final point in time from the specified target

    Arguments
    ---------
    variables: List
        List of time-indexed variables that will be penalized
    time_set: Set
        Time set that indexes the provided variables. Penalties are applied
        at the last point in this set.
    target_data: ~scalar_data.ScalarData
        ScalarData object containing the target for (at least) the variables
        to be penalized
    weight_data: ~scalar_data.ScalarData (optional)
        ScalarData object containing the penalty weights for (at least) the
        variables to be penalized
    variable_set: Set (optional)
        Set indexing the list of variables provided, if such a set already
        exists

    Returns
    -------
    Set, Expression
        Set indexing the list of variables provided and an Expression,
        indexed by this set, containing the weighted penalty expressions

    """
    t = time_set.last()
    return get_penalty_at_time(
        variables,
        t,
        target_data,
        weight_data=weight_data,
        time_set=time_set,
        variable_set=variable_set,
    )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Test the Pyomo BAR writer
#
import itertools
import re
import glob
import os
from os.path import abspath, dirname, join

currdir = dirname(abspath(__file__)) + os.sep
datadir = abspath(join(currdir, "..", "ampl")) + os.sep

import pyomo.common.unittest as unittest
import pyomo.common

import pyomo.scripting.pyomo_main as main

parameterized, param_available = pyomo.common.dependencies.attempt_import(
    'parameterized'
)
if not param_available:
    raise unittest.SkipTest('Parameterized is not available.')

names = []
# add test methods to classes
for f in itertools.chain(
    glob.glob(join(datadir, '*_testCase.py')), glob.glob(join(currdir, '*_testCase.py'))
):
    names.append(re.split('[._]', os.path.basename(f))[0])


class Tests(unittest.TestCase):
    def pyomo(self, cmd):
        os.chdir(currdir)
        output = main.main(['convert', '--logging=quiet', '-c'] + cmd)
        return output


class BaselineTests(Tests):
    def __init__(self, *args, **kwds):
        Tests.__init__(self, *args, **kwds)

    #
    # The following test generates an BAR file for the test case
    # and checks that it matches the current pyomo baseline BAR file
    #
    @parameterized.parameterized.expand(input=names)
    def barwriter_baseline_test(self, name):
        baseline = join(currdir, name + '.pyomo.bar')
        output = join(currdir, name + '.test.bar')
        if not os.path.exists(baseline):
            self.skipTest("baseline file (%s) not found" % (baseline,))

        if os.path.exists(datadir + name + '_testCase.py'):
            testDir = datadir
        else:
            testDir = currdir
        testCase = testDir + name + '_testCase.py'

        if os.path.exists(testDir + name + '.dat'):
            self.pyomo(['--output=' + output, testCase, testDir + name + '.dat'])
        else:
            self.pyomo(['--output=' + output, testCase])

        # Check that the pyomo BAR file matches its own baseline
        with open(baseline, 'r') as f1, open(output, 'r') as f2:
            f1_contents = list(filter(None, f1.read().split()))
            f2_contents = list(filter(None, f2.read().split()))
            for item1, item2 in itertools.zip_longest(f1_contents, f2_contents):
                try:
                    self.assertAlmostEqual(float(item1), float(item2))
                except:
                    self.assertEqual(
                        item1,
                        item2,
                        "\n\nbaseline: %s\ntestFile: %s\n" % (baseline, output),
                    )
        os.remove(join(currdir, name + '.test.bar'))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
import pyomo.common.unittest as unittest

from pyomo.common.dependencies import (
    numpy as np,
    numpy_available,
    pandas as pd,
    pandas_available,
)

from pyomo.environ import (
    ConcreteModel,
    Var,
    RangeSet,
    Param,
    Objective,
    Set,
    Constraint,
    Reals,
)
from pyomo.core.expr import MonomialTermExpression
from pyomo.core.expr.ndarray import NumericNDArray
from pyomo.core.expr.numvalue import as_numeric
from pyomo.core.expr.compare import compare_expressions
from pyomo.core.expr.relational_expr import InequalityExpression

from pyomo.repn import generate_standard_repn


@unittest.skipUnless(numpy_available, 'numpy is not available')
class TestNumPy(unittest.TestCase):
    def test_numpy_scalar_times_scalar_var(self):
        # Test issue #685
        m = ConcreteModel()
        m.x = Var()
        e = np.float64(5) * m.x
        self.assertIs(type(e), MonomialTermExpression)
        self.assertTrue(compare_expressions(e, 5.0 * m.x))

        e = m.x * np.float64(5)
        self.assertIs(type(e), MonomialTermExpression)
        self.assertTrue(compare_expressions(e, 5.0 * m.x))

    def test_initialize_param_from_ndarray(self):
        # Test issue #611
        samples = 10
        c1 = 0.5
        c2 = 0.5

        model = ConcreteModel()
        model.i = RangeSet(samples)

        def init_x(model, i):
            return np.random.rand(1)

        def init_y(model, i):
            return c1 * (model.x[i] ** 2) + c2 * model.x[i]

        model.x = Param(model.i, initialize=init_x)
        model.y = Param(model.i, initialize=init_y, domain=Reals)
        model.c_1 = Var(initialize=1)
        model.c_2 = Var(initialize=1)
        model.error = Objective(
            # Sum squared error of quadratic fit
            expr=sum(
                (model.c_1 * model.x[i] ** 2 + model.c_2 * model.x[i] - model.y[i]) ** 2
                for i in model.i
            )
        )
        # model.pprint()

        repn = generate_standard_repn(model.error.expr, compute_values=True)
        self.assertIsNone(repn.nonlinear_expr)
        self.assertEqual(len(repn.quadratic_vars), 3)
        for i in range(3):
            self.assertGreater(repn.quadratic_coefs[i], 0)
        self.assertEqual(len(repn.linear_vars), 2)
        for i in range(2):
            self.assertLess(repn.linear_coefs[i], 0)
        self.assertGreater(repn.constant, 0)

    def test_create_objective_from_numpy(self):
        # Test issue #87
        model = ConcreteModel()

        nsample = 3
        nvariables = 2
        X0 = np.array(range(nsample)).reshape([nsample, 1])
        model.X = 1 + np.array(range(nsample * nvariables)).reshape(
            (nsample, nvariables)
        )
        X = np.concatenate([X0, model.X], axis=1)

        model.I = RangeSet(1, nsample)
        model.J = RangeSet(1, nvariables)

        error = np.ones((nsample, 1))
        beta = np.ones((nvariables + 1, 1))
        model.Y = np.dot(X, beta) + error

        model.beta = Var(model.J)
        model.beta0 = Var()

        def obj_fun(model):
            return sum(
                abs(
                    model.Y[i - 1]
                    - (
                        model.beta0
                        + sum(model.X[i - 1, j - 1] * model.beta[j] for j in model.J)
                    )
                )
                for i in model.I
            )

        model.OBJ = Objective(rule=obj_fun)

        def obj_fun_quad(model):
            return sum(
                (
                    model.Y[i - 1]
                    - (
                        model.beta0
                        + sum(model.X[i - 1, j - 1] * model.beta[j] for j in model.J)
                    )
                )
                ** 2
                for i in model.I
            )

        model.OBJ_QUAD = Objective(rule=obj_fun_quad)

        self.assertEqual(
            str(model.OBJ.expr),
            "abs(4.0 - (beta[1] + 2*beta[2] + beta0)) + "
            "abs(9.0 - (3*beta[1] + 4*beta[2] + beta0)) + "
            "abs(14.0 - (5*beta[1] + 6*beta[2] + beta0))",
        )
        self.assertEqual(model.OBJ.expr.polynomial_degree(), None)
        self.assertEqual(model.OBJ_QUAD.expr.polynomial_degree(), 2)

    @unittest.skipUnless(pandas_available, "pandas is not available")
    def test_param_from_pandas(self):
        # Test issue #68
        model = ConcreteModel()
        model.I = Set(initialize=range(6))

        model.P0 = Param(
            model.I, initialize={0: 400.0, 1: 0.0, 2: 0.0, 3: 0.0, 4: 0.0, 5: 240.0}
        )
        model.P1 = Param(
            model.I,
            initialize=pd.Series(
                {0: 400.0, 1: 0.0, 2: 0.0, 3: 0.0, 4: 0.0, 5: 240.0}
            ).to_dict(),
        )
        model.P2 = Param(
            model.I,
            initialize=pd.Series({0: 400.0, 1: 0.0, 2: 0.0, 3: 0.0, 4: 0.0, 5: 240.0}),
        )

        # model.pprint()
        self.assertEqual(list(model.P0.values()), list(model.P1.values()))
        self.assertEqual(list(model.P0.values()), list(model.P2.values()))

        model.V = Var(model.I, initialize=0)

        def rule(m, l):
            return -m.P0[l] <= m.V[l]

        model.Constraint0 = Constraint(model.I, rule=rule)

        def rule(m, l):
            return -m.P1[l] <= m.V[l]

        model.Constraint1 = Constraint(model.I, rule=rule)

        def rule(m, l):
            return -m.P2[l] <= m.V[l]

        model.Constraint2 = Constraint(model.I, rule=rule)

        # TODO: support vector operations between Indexed objects
        # model.Constraint0a = Constraint(model.I, rule=model.P0 <= model.V)
        # model.Constraint1a = Constraint(model.I, rule=model.P1 <= model.V)
        # model.Constraint2a = Constraint(model.I, rule=model.P2 <= model.V)

    @unittest.skipUnless(pandas_available, "pandas is not available")
    def test_param_from_pandas_series_index(self):
        m = ConcreteModel()
        s = pd.Series([1, 3, 5], index=['T1', 'T2', 'T3'])

        # Params treat Series as maps (so the Series index matters)
        m.I = Set(initialize=s.index)
        m.p1 = Param(m.I, initialize=s)
        self.assertEqual(m.p1.extract_values(), {'T1': 1, 'T2': 3, 'T3': 5})
        m.p2 = Param(s.index, initialize=s)
        self.assertEqual(m.p2.extract_values(), {'T1': 1, 'T2': 3, 'T3': 5})
        with self.assertRaisesRegex(
            KeyError, "Index 'T1' is not valid for indexed component 'p3'"
        ):
            m.p3 = Param([0, 1, 2], initialize=s)

        # Sets treat Series as lists
        m.J = Set(initialize=s)
        self.assertEqual(set(m.J), {1, 3, 5})

    def test_numpy_float(self):
        # Test issue #31
        m = ConcreteModel()

        m.T = Set(initialize=range(3))
        m.v = Var(initialize=1, bounds=(0, None))
        m.c = Var(m.T, initialize=20)
        h = [np.float32(1.0), 1.0, 1]

        def rule(m, t):
            return m.c[0] == h[t] * m.c[0]

        m.x = Constraint(m.T, rule=rule)

        def rule(m, t):
            return m.c[0] == h[t] * m.c[0] * m.v

        m.y = Constraint(m.T, rule=rule)

        def rule(m, t):
            return m.c[0] == h[t] * m.v

        m.z = Constraint(m.T, rule=rule)

        # m.pprint()
        for t in m.T:
            self.assertTrue(compare_expressions(m.x[0].expr, m.x[t].expr))
            self.assertTrue(compare_expressions(m.y[0].expr, m.y[t].expr))
            self.assertTrue(compare_expressions(m.z[0].expr, m.z[t].expr))

    def test_indexed_constraint(self):
        m = ConcreteModel()
        m.x = Var([0, 1, 2, 3])
        A = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        b = np.array([10, 20])
        m.c = Constraint([0, 1], expr=A @ m.x <= b)
        self.assertTrue(
            compare_expressions(
                m.c[0].expr, m.x[0] + 2 * m.x[1] + 3 * m.x[2] + 4 * m.x[3] <= 10
            )
        )
        self.assertTrue(
            compare_expressions(
                m.c[1].expr, 5 * m.x[0] + 6 * m.x[1] + 7 * m.x[2] + 8 * m.x[3] <= 20
            )
        )

    def test_numpy_array_copy_errors(self):
        # Defer testing until here to avoid the unconditional numpy dereference
        if int(np.__version__.split('.')[0]) < 2:
            self.skipTest("requires numpy>=2")

        m = ConcreteModel()
        m.x = Var([0, 1, 2])
        with self.assertRaisesRegex(
            ValueError,
            "Pyomo IndexedComponents do not support conversion to NumPy "
            "arrays without generating a new array",
        ):
            np.asarray(m.x, copy=False)

    def test_numpy_array_dtype_errors(self):
        m = ConcreteModel()
        m.x = Var([0, 1, 2])
        # object is OK
        a = np.asarray(m.x, object)
        self.assertEqual(a.shape, (3,))
        # None is OK
        a = np.asarray(m.x, None)
        self.assertEqual(a.shape, (3,))
        # Anything else is an error
        with self.assertRaisesRegex(
            ValueError,
            "Pyomo IndexedComponents can only be converted to NumPy arrays "
            r"with dtype=object \(received dtype=.*int32",
        ):
            a = np.asarray(m.x, np.int32)

    def test_init_param_from_ndarray(self):
        # Test issue #2033
        m = ConcreteModel()
        m.ix_set = RangeSet(2)

        p_init = np.array([0, 5])

        def init_workaround(model, i):
            return p_init[i - 1]

        m.p = Param(m.ix_set, initialize=init_workaround)
        m.v = Var(m.ix_set)
        expr = m.p[1] > m.v[1]
        self.assertIsInstance(expr, InequalityExpression)
        self.assertEqual(str(expr), "v[1]  <  0")
        expr = m.p[2] > m.v[2]
        self.assertIsInstance(expr, InequalityExpression)
        self.assertEqual(str(expr), "v[2]  <  5")


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Test the Pyomo command-line interface
#

from itertools import zip_longest
import json
import re
import os
import sys
from os.path import abspath, dirname, join
from filecmp import cmp
import subprocess
import pyomo.common.unittest as unittest

from pyomo.common.dependencies import yaml_available
from pyomo.common.fileutils import this_file_dir
from pyomo.common.tee import capture_output
from pyomo.common.tempfiles import TempfileManager
import pyomo.core
import pyomo.scripting.pyomo_main as main
from pyomo.opt import check_available_solvers

from io import StringIO

currdir = this_file_dir()

_diff_tol = 1e-6

deleteFiles = True

solvers = None


class BaseTester(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        global solvers
        import pyomo.environ

        solvers = check_available_solvers('glpk')

    def pyomo(self, cmd, **kwds):
        if 'root' in kwds:
            OUTPUT = kwds['root'] + '.out'
            results = kwds['root'] + '.jsn'
            TempfileManager.add_tempfile(OUTPUT, exists=False)
            TempfileManager.add_tempfile(results, exists=False)
        else:
            OUTPUT = StringIO()
            results = 'results.jsn'
            TempfileManager.create_tempfile(suffix='results.jsn')
        with capture_output(OUTPUT):
            try:
                _dir = os.getcwd()
                os.chdir(currdir)
                args = [
                    'solve',
                    '--solver=glpk',
                    '--results-format=json',
                    '--save-results=%s' % results,
                ]
                if type(cmd) is list:
                    args.extend(cmd)
                elif cmd.endswith('json') or cmd.endswith('yaml'):
                    args.append(cmd)
                else:
                    args.extend(re.split('[ ]+', cmd))
                output = main.main(args)
            finally:
                os.chdir(_dir)
        if not 'root' in kwds:
            return OUTPUT.getvalue()
        return output

    def setUp(self):
        if not 'glpk' in solvers:
            self.skipTest("GLPK is not installed")
        TempfileManager.push()

    def tearDown(self):
        TempfileManager.pop(remove=deleteFiles or self.currentTestPassed())

    def run_pyomo(self, cmd, root):
        results = root + '.jsn'
        TempfileManager.add_tempfile(results, exists=False)
        output = root + '.out'
        TempfileManager.add_tempfile(output, exists=False)
        cmd = [
            'pyomo',
            'solve',
            '--solver=glpk',
            '--results-format=json',
            '--save-results=%s' % results,
        ] + cmd
        with open(output, 'w') as f:
            result = subprocess.run(cmd, stdout=f, stderr=f)
        return result


class TestJson(BaseTester):
    def compare_json(self, file1, file2):
        with open(file1, 'r') as f1, open(file2, 'r') as f2:
            f1_contents = json.load(f1)
            f2_contents = json.load(f2)
            self.assertStructuredAlmostEqual(
                f2_contents, f1_contents, abstol=_diff_tol, allow_second_superset=True
            )

    def filter_items(self, items):
        filtered = []
        for i in items:
            if not (i.startswith('/') or i.startswith(":\\", 1)):
                try:
                    filtered.append(float(i))
                except:
                    filtered.append(i)
        return filtered

    def compare_files(self, file1, file2):
        try:
            self.assertTrue(
                cmp(file1, file2), msg="Files %s and %s differ" % (file1, file2)
            )
        except:
            with open(file1, 'r') as f1, open(file2, 'r') as f2:
                f1_contents = f1.read().strip().split('\n')
                f2_contents = f2.read().strip().split('\n')
                f1_filtered = []
                f2_filtered = []
                for item1, item2 in zip_longest(f1_contents, f2_contents):
                    if not item1:
                        f1_filtered.append(item1)
                    elif not item1.startswith('['):
                        items1 = item1.strip().split()
                        f1_filtered.append(self.filter_items(items1))
                    if not item2:
                        f2_filtered.append(item2)
                    elif not item2.startswith('['):
                        items2 = item2.strip().split()
                        f2_filtered.append(self.filter_items(items2))
                self.assertStructuredAlmostEqual(
                    f2_filtered, f1_filtered, abstol=1e-6, allow_second_superset=True
                )

    def test1_simple_pyomo_execution(self):
        # Simple execution of 'pyomo'
        self.pyomo(
            [join(currdir, 'pmedian.py'), join(currdir, 'pmedian.dat')],
            root=join(currdir, 'test1'),
        )
        self.compare_json(join(currdir, 'test1.jsn'), join(currdir, 'test1.txt'))

    def test1a_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' in a subprocess
        files = [
            os.path.join(currdir, 'pmedian.py'),
            os.path.join(currdir, 'pmedian.dat'),
        ]
        self.run_pyomo(files, root=os.path.join(currdir, 'test1a'))
        self.compare_json(join(currdir, 'test1a.jsn'), join(currdir, 'test1.txt'))

    def test1b_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with a configuration file
        self.pyomo(join(currdir, 'test1b.json'), root=join(currdir, 'test1'))
        self.compare_json(join(currdir, 'test1.jsn'), join(currdir, 'test1.txt'))

    def test2_bad_model_name(self):
        # Run pyomo with bad --model-name option value
        self.pyomo(
            '--model-name=dummy pmedian.py pmedian.dat', root=join(currdir, 'test2')
        )
        self.compare_files(join(currdir, "test2.out"), join(currdir, "test2.txt"))

    def test2b_bad_model_name(self):
        # Run pyomo with bad --model-name option value (configfile)
        self.pyomo(join(currdir, 'test2b.json'), root=join(currdir, 'test2'))
        self.compare_files(join(currdir, "test2.out"), join(currdir, "test2.txt"))

    def test3_missing_model_object(self):
        # Run pyomo with model that does not define model object
        self.pyomo('pmedian1.py pmedian.dat', root=join(currdir, 'test3'))
        self.compare_json(join(currdir, "test3.jsn"), join(currdir, "test1.txt"))

    def test4_valid_modelname_option(self):
        # Run pyomo with good --model-name option value
        self.pyomo(
            '--model-name=MODEL ' + join(currdir, 'pmedian1.py pmedian.dat'),
            root=join(currdir, 'test4'),
        )
        self.compare_json(join(currdir, "test4.jsn"), join(currdir, "test1.txt"))

    def test4b_valid_modelname_option(self):
        # Run pyomo with good 'object name' option value (configfile)
        self.pyomo(join(currdir, 'test4b.json'), root=join(currdir, 'test4b'))
        self.compare_json(join(currdir, "test4b.jsn"), join(currdir, "test1.txt"))

    def test5_create_model_fcn(self):
        # """Run pyomo with create_model function"""
        self.pyomo('pmedian2.py pmedian.dat', root=join(currdir, 'test5'))
        self.compare_files(join(currdir, "test5.out"), join(currdir, "test5.txt"))

    def test5b_create_model_fcn(self):
        # Run pyomo with create_model function (configfile)
        self.pyomo(join(currdir, 'test5b.json'), root=join(currdir, 'test5'))
        self.compare_files(join(currdir, "test5.out"), join(currdir, "test5.txt"))

    def test8_instanceonly_option(self):
        # """Run pyomo with --instance-only option"""
        output = self.pyomo(
            '--instance-only pmedian.py pmedian.dat', root=join(currdir, 'test8')
        )
        self.assertEqual(type(output.retval.instance), pyomo.core.ConcreteModel)
        # Check that the results file was NOT created
        self.assertFalse(os.path.exists(join(currdir, 'test8.jsn')))

    def test8b_instanceonly_option(self):
        # Run pyomo with --instance-only option (configfile)
        output = self.pyomo(join(currdir, 'test8b.json'), root=join(currdir, 'test8'))
        self.assertEqual(type(output.retval.instance), pyomo.core.ConcreteModel)
        # Check that the results file was NOT created
        self.assertFalse(os.path.exists(join(currdir, 'test8.jsn')))

    def test9_disablegc_option(self):
        # """Run pyomo with --disable-gc option"""
        output = self.pyomo(
            '--disable-gc pmedian.py pmedian.dat', root=join(currdir, 'test9')
        )
        self.assertEqual(type(output.retval.instance), pyomo.core.ConcreteModel)

    def test9b_disablegc_option(self):
        # Run pyomo with --disable-gc option (configfile)
        output = self.pyomo(join(currdir, 'test9b.json'), root=join(currdir, 'test9'))
        self.assertEqual(type(output.retval.instance), pyomo.core.ConcreteModel)

    def test12_output_option(self):
        # """Run pyomo with --output option"""
        log = join(currdir, 'test12.log')
        TempfileManager.add_tempfile(log, exists=False)
        self.pyomo(
            '--logfile=%s pmedian.py pmedian.dat' % (log,), root=join(currdir, 'test12')
        )
        self.compare_json(join(currdir, "test12.jsn"), join(currdir, "test12.txt"))

    def test12b_output_option(self):
        # Run pyomo with --output option (configfile)
        log = join(currdir, 'test12b.log')
        TempfileManager.add_tempfile(log, exists=False)
        self.pyomo(join(currdir, 'test12b.json'), root=join(currdir, 'test12'))
        self.compare_json(join(currdir, "test12.jsn"), join(currdir, "test12.txt"))

    def test14_concrete_model_with_constraintlist(self):
        # Simple execution of 'pyomo' with a concrete model and constraint lists
        self.pyomo('pmedian4.py', root=join(currdir, 'test14'))
        self.compare_json(join(currdir, "test14.jsn"), join(currdir, "test14.txt"))

    def test14b_concrete_model_with_constraintlist(self):
        # Simple execution of 'pyomo' with a concrete model and constraint lists (configfile)
        self.pyomo('pmedian4.py', root=join(currdir, 'test14'))
        self.compare_json(join(currdir, "test14.jsn"), join(currdir, "test14.txt"))

    def test15_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with options
        self.pyomo(
            [
                '--solver-options="mipgap=0.02 cuts="',
                join(currdir, 'pmedian.py'),
                'pmedian.dat',
            ],
            root=join(currdir, 'test15'),
        )
        self.compare_json(join(currdir, "test15.jsn"), join(currdir, "test1.txt"))

    def test15b_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with options
        self.pyomo(join(currdir, 'test15b.json'), root=join(currdir, 'test15b'))
        self.compare_json(join(currdir, "test15b.jsn"), join(currdir, "test1.txt"))

    def test15c_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with options
        self.pyomo(join(currdir, 'test15c.json'), root=join(currdir, 'test15c'))
        self.compare_json(join(currdir, "test15c.jsn"), join(currdir, "test1.txt"))


@unittest.skipIf(not yaml_available, "YAML not available available")
class TestWithYaml(BaseTester):
    def compare_json(self, file1, file2):
        with open(file1, 'r') as f1, open(file2, 'r') as f2:
            f1_contents = json.load(f1)
            f2_contents = json.load(f2)
            self.assertStructuredAlmostEqual(
                f2_contents, f1_contents, abstol=_diff_tol, allow_second_superset=True
            )

    def test15b_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with options
        self.pyomo(join(currdir, 'test15b.yaml'), root=join(currdir, 'test15b'))
        self.compare_json(join(currdir, "test15b.jsn"), join(currdir, "test1.txt"))

    def test15c_simple_pyomo_execution(self):
        # Simple execution of 'pyomo' with options
        self.pyomo(join(currdir, 'test15c.yaml'), root=join(currdir, 'test15c'))
        self.compare_json(join(currdir, "test15c.jsn"), join(currdir, "test1.txt"))


if __name__ == "__main__":
    deleteFiles = False
    unittest.main()
